# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
from iotics.lib.identity.api.advanced_api import *  # noqa: F403,F401
from iotics.lib.identity.api.high_level_api import *  # noqa: F403,F401
from iotics.lib.identity.api.regular_api import *  # noqa: F403,F401



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.


class IdentityBaseException(Exception):
    """
    Identity base exception. All the exceptions from the iotic.lib.identity extends this exception.
    """


class IdentityDependencyError(IdentityBaseException):
    """
    Raised when an unexpected error is raised by a dependency due to a version incompatibility.
    """


class IdentityValidationError(IdentityBaseException):
    """
    Raised when a user input is invalid.
    """


class IdentityInvalidDocumentError(IdentityValidationError):
    """
    Raised when a RegisterDocument is invalid.
    """


class IdentityInvalidProofError(IdentityInvalidDocumentError):
    """
    Raised when a Proof (for RegisterDocument, Delegation, ...) is invalid.
    """


class IdentityInvalidDocumentDelegationError(IdentityInvalidDocumentError):
    """
    Raised when a RegisterDocument delegation (authentication or control) is invalid.
    """


class IdentityRegisterDocumentKeyConflictError(IdentityInvalidDocumentError):
    """
    Raised when a RegisteredDocument key name is not unique.
    """


class IdentityRegisterDocumentKeyNotFoundError(IdentityInvalidDocumentError):
    """
    Raised when a register document key is not found.
    """


class IdentityRegisterIssuerNotFoundError(IdentityInvalidDocumentError):
    """
    Raised when a RegisteredDocument issuer is not found.
    """


class IdentityInvalidRegisterIssuerError(IdentityInvalidDocumentError):
    """
    Raised when a RegisteredDocument issuer is invalid.
    """


class IdentityResolverError(IdentityBaseException):
    """
    Raised when an error occurs while interacting with the resolver.
    """


class IdentityResolverConflictError(IdentityResolverError):
    """
    Raised when a register document already exist with a different owners
    """


class IdentityResolverDocNotFoundError(IdentityResolverError):
    """
    Raised when a RegisterDocument is not found against the resolver.
    """


class IdentityResolverCommunicationError(IdentityResolverError):
    """
    Raised when a communication error occurs while interacting with the resolver.
    """


class IdentityResolverTimeoutError(IdentityResolverError):
    """
    Raised when a timeout error occurs while interacting with the resolver.
    """


class IdentityResolverHttpError(IdentityResolverError):
    """
    Raised when a HTTPError occurs while interacting with the REST resolver.
    """


class IdentityResolverHttpDocNotFoundError(IdentityResolverHttpError, IdentityResolverDocNotFoundError):
    """
    Raised when a RegisterDocument is not found against the REST resolver.
    """


class IdentityAuthenticationFailed(IdentityBaseException):
    """
    Raised when verify authentication fails.
    """


class IdentityNotAllowed(IdentityAuthenticationFailed):
    """
    Raised when identity not allowed for authentication or control.
    """



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from dataclasses import dataclass

from iotics.lib.identity.const import ISSUER_SEPARATOR
from iotics.lib.identity.error import IdentityValidationError
from iotics.lib.identity.validation.identity import IdentityValidation


@dataclass(frozen=True)
class Issuer:
    did: str
    name: str

    @staticmethod
    def build(did: str, name: str) -> 'Issuer':
        """
        Build a valid issuer.
        :param did: issuer decentralised identifier
        :param name: issuer name
        :return: valid issuer

        :raises:
            IdentityValidationError: if invalid name or did

        """
        IdentityValidation.validate_identifier(did)
        IdentityValidation.validate_key_name(name)
        return Issuer(did, name)

    @staticmethod
    def from_string(issuer_string: str) -> 'Issuer':
        """
        Build a valid issuer from issuer string.
        :param issuer_string: issuer string
        :return: valid issuer

        :raises:
            IdentityValidationError: if invalid issuer string
        """
        parts = issuer_string.split(ISSUER_SEPARATOR)
        if len(parts) != 2:
            raise IdentityValidationError(
                f'Invalid issuer string {issuer_string} should be of the form of [did]{ISSUER_SEPARATOR}[name]')
        return Issuer.build(parts[0], f'{ISSUER_SEPARATOR}{parts[1]}')

    def __str__(self) -> str:
        return f'{self.did}{self.name}'


@dataclass(frozen=True)
class IssuerKey:
    issuer: Issuer
    public_key_base58: str

    @staticmethod
    def build(did: str, name: str, public_key_base58: str) -> 'IssuerKey':
        """
        Build an issuer key from identifier, name and public key.
        :param did: issuer decentralised identifier
        :param name: issuer name
        :param public_key_base58: public key base58
        :return: issuer key with valid issuer

        :raises:
            IdentityValidationError: if invalid name or did
        """
        return IssuerKey(Issuer.build(did, name), public_key_base58)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from abc import ABC, abstractmethod

from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.register.document import RegisterDocument


class ResolverClient(ABC):

    @abstractmethod
    def get_document(self, doc_id: str) -> RegisterDocument:
        """
        Get a valid register document from the resolver.
        :param doc_id: register document decentralised identifier
        :return: valid register document

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverDocNotFoundError: if document not found
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        raise NotImplementedError

    @abstractmethod
    def register_document(self, document: RegisterDocument, private_key: ec.EllipticCurvePrivateKey,
                          issuer: Issuer, audience: str = ''):
        """
        Register a register document against the Resolver.
        :param document: register document
        :param issuer: issuer
        :param private_key: issuer private key
        :param audience: audience

        :raises:
            IdentityResolverError: if resolver error
            IdentityResolverError: if can not serialize document
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        raise NotImplementedError



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

import base64
from dataclasses import dataclass

import jwt
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.const import TOKEN_ALGORITHM
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.jwt import JwtTokenHelper
from iotics.lib.identity.crypto.key_pair_secrets import KeyPairSecrets, KeyPairSecretsHelper
from iotics.lib.identity.error import IdentityInvalidRegisterIssuerError, IdentityValidationError
from iotics.lib.identity.register.document_helper import RegisterDocumentHelper
from iotics.lib.identity.register.resolver import ResolverClient


@dataclass(frozen=True)
class Proof:
    issuer: Issuer
    content: bytes
    signature: str

    @staticmethod
    def from_challenge_token(resolver_client: ResolverClient, challenge_token: str) -> 'Proof':
        """
        Build proof from challenge token.
        :param resolver_client: resolver client to get the registered documents
        :param challenge_token: jwt challenge token
        :return: valid proof

        :raises:
            IdentityValidationError: if invalid challenge token
        """
        decoded_token = JwtTokenHelper.decode_token(challenge_token)
        iss = decoded_token.get('iss')
        aud = decoded_token.get('aud')
        if not iss or not aud:
            raise IdentityValidationError('Invalid challenge token, missing \'iss\' or \'aud\'')

        issuer = Issuer.from_string(iss)
        doc = resolver_client.get_document(issuer.did)
        get_controller_doc = resolver_client.get_document
        issuer_key = RegisterDocumentHelper.get_valid_issuer_key_for_control_only(doc, issuer.name, get_controller_doc)
        if not issuer_key:
            raise IdentityInvalidRegisterIssuerError(f'Invalid issuer {issuer}')
        verified_token = JwtTokenHelper.decode_and_verify_token(challenge_token, issuer_key.public_key_base58, aud)
        return Proof(issuer_key.issuer, aud.encode('ascii'), verified_token['proof'])

    @staticmethod
    def build(key_pair: KeyPairSecrets, issuer: Issuer, content: bytes) -> 'Proof':
        """
        Build a proof.
        :param key_pair: secrets used to build the proof signature
        :param issuer: proof issuer
        :param content:  proof content
        :return: proof

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityDependencyError: if incompatible library dependency
        """
        private_key = KeyPairSecretsHelper.get_private_key(key_pair)
        sig = private_key.sign(content, ec.ECDSA(hashes.SHA256()))
        proof = base64.b64encode(sig).decode('ascii')
        return Proof(issuer=issuer, content=content, signature=proof)


def build_new_challenge_token(proof: Proof, private_key: ec.EllipticCurvePrivateKey) -> str:
    """
    Build a new challenge token from a proof.
    :param proof: proof
    :param private_key: private key
    :return: jwt challenge token

    :raises:
        IdentityValidationError: if can not encode the token

    """
    try:
        return jwt.encode({'iss': str(proof.issuer), 'aud': proof.content.decode('ascii'), 'proof': proof.signature},
                          private_key, algorithm=TOKEN_ALGORITHM)  # type: ignore
    except (TypeError, ValueError) as err:
        raise IdentityValidationError(f'Can not create challenge token for {proof}: \'{err}\'') from err



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
# For pkgutil namespace compatibility only. Must NOT contain anything else. See also:
# https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
__path__ = __import__('pkgutil').extend_path(__path__, __name__)  # type: ignore



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from dataclasses import dataclass
from typing import Tuple

import base58
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.error import IdentityDependencyError, IdentityValidationError


@dataclass(frozen=True)
class KeyPair:
    private_key: ec.EllipticCurvePrivateKey
    public_bytes: bytes
    public_base58: str


class KeysHelper:

    @staticmethod
    def get_private_ECDSA(private_expo: str) -> ec.EllipticCurvePrivateKey:
        """
        Get private key (ECDSA) from private exponent
        :param private_expo: private exponent as hex string
        :return: private ECDSA key

        :raises:
            IdentityDependencyError: if incompatible EllipticCurve dependency
        """
        sbin = bytes.fromhex(private_expo)
        sint = int.from_bytes(sbin, 'big', signed=False)

        try:
            return ec.derive_private_key(sint, ec.SECP256K1(), default_backend())
        except Exception as err:
            raise IdentityDependencyError(f'Dependency cryptography failed to derive private key: {err}') from err

    @staticmethod
    def get_public_keys_from_private_ECDSA(private_key: ec.EllipticCurvePrivateKey) -> Tuple[bytes, str]:
        """
        Get public keys (bytes and base58) from private key (ECDSA)
        :param private_key: private key
        :return: public key bytes, public key base58
        """
        public_key = private_key.public_key()

        public_bytes = public_key.public_bytes(encoding=serialization.Encoding.X962,
                                               format=serialization.PublicFormat.UncompressedPoint)
        return public_bytes, base58.b58encode(public_bytes).decode('ascii')

    @staticmethod
    def get_public_ECDSA_from_base58(public_base58: str) -> ec.EllipticCurvePublicKey:
        """
        Get public key ECDSA from public key base58
        :param public_base58: public key base58
        :return: public key ECDSA

        :raises:
            IdentityValidationError: if invalid public base58 key
        """
        try:
            public_bytes = base58.b58decode(public_base58)
            public_ecdsa = ec.EllipticCurvePublicKey.from_encoded_point(ec.SECP256K1(), public_bytes)
            return public_ecdsa
        except ValueError as err:
            raise IdentityValidationError(f'Can not convert public key base58 to ECDSA: \'{err}\'') from err



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from typing import Optional, Union

from iotics.lib.identity.api.advanced_api import AdvancedIdentityLocalApi, AdvancedIdentityRegisterApi
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.key_pair_secrets import build_agent_secrets, build_twin_secrets, build_user_secrets, \
    KeyPairSecrets, KeyPairSecretsHelper, SeedMethod
from iotics.lib.identity.crypto.keys import KeyPair
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.key_pair import RegisteredIdentity
from iotics.lib.identity.register.rest_resolver import get_rest_resolver_client


class IdentityApi:
    def __init__(self, advanced_api: AdvancedIdentityRegisterApi):
        self.advanced_api = advanced_api

    def create_user_identity(self, user_seed: bytes, user_key_name: str, user_name: str,
                             seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                             password: str = '', override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a user identity.
        :param user_seed: user seed (secrets)
        :param user_key_name: user key name (secrets)
        :param password: Optional user password (secrets)
        :param user_name: Optional user name (default #user-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param override_doc: override registered identity document if already exist (default False)
        :return: user registered identity

        :raises:
            IdentityValidationError: if invalid secrets or name
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair_secrets = build_user_secrets(user_seed, user_key_name, seed_method, password)
        return self.advanced_api.new_registered_user_identity(key_pair_secrets, user_name, override_doc)

    def create_agent_identity(self, agent_seed: bytes, agent_key_name: str, agent_name: str,
                              seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                              password: str = '', override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a agent identity.
        :param agent_seed: agent seed (secrets)
        :param agent_key_name: agent key name (secrets)
        :param password: Optional agent password (secrets)
        :param agent_name: Optional agent name (default #agent-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param override_doc: override registered identity document if already exist (default False)
        :return: agent registered identity

        :raises:
            IdentityValidationError: if invalid secrets or name
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair_secrets = build_agent_secrets(agent_seed, agent_key_name, seed_method, password)
        return self.advanced_api.new_registered_agent_identity(key_pair_secrets, agent_name, override_doc)

    def create_twin_identity(self, twin_seed: bytes, twin_key_name: str, twin_name: str,
                             seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                             password: str = '', override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a twin identity.
        :param twin_seed: twin seed (secrets)
        :param twin_key_name: twin key name (secrets)
        :param password: Optional twin password (secrets)
        :param twin_name: Optional twin name (default #twin-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param override_doc: override registered identity document if already exist (default False)
        :return: twin registered identity

        :raises:
            IdentityValidationError: if invalid secrets or name
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair_secrets = build_twin_secrets(twin_seed, twin_key_name, seed_method, password)
        return self.advanced_api.new_registered_twin_identity(key_pair_secrets, twin_name, override_doc)

    @staticmethod
    def _get_identity(key_pair_secrets: KeyPairSecrets, did: str, name: str) -> RegisteredIdentity:
        return RegisteredIdentity(key_pair_secrets=key_pair_secrets,
                                  issuer=Issuer.build(did, name))

    def get_user_identity(self, key_seed: bytes, key_name: str, user_did: str, user_name: str,
                          seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                          password: str = '') -> RegisteredIdentity:
        """
        Get user registered identity from secrets.
        :param key_seed: user secrets
        :param key_name: user key name
        :param user_did: user decentralised identifier
        :param user_name: user name
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param password: Optional user password (secrets)
        :return: user registered identity

        :raises:
            IdentityValidationError: if invalid user name
        """
        key_pair_secrets = build_user_secrets(key_seed, key_name, seed_method, password)
        return self._get_identity(key_pair_secrets, user_did, user_name)

    def get_agent_identity(self, key_seed: bytes, key_name: str, agent_did: str, agent_name: str,
                           seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                           password: str = '') -> RegisteredIdentity:
        """
        Get agent registered identity from secrets.
        :param key_seed: agent secrets
        :param key_name: agent key name
        :param agent_did: agent decentralised identifier
        :param agent_name: agent name
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param password: Optional agent password (secrets)
        :return: agent registered identity

        :raises:
            IdentityValidationError: if invalid agent name
        """
        key_pair_secrets = build_agent_secrets(key_seed, key_name, seed_method, password)
        return self._get_identity(key_pair_secrets, agent_did, agent_name)

    def get_twin_identity(self, key_seed: bytes, key_name: str, twin_did: str, twin_name: str,
                          seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                          password: str = '') -> RegisteredIdentity:
        """
        Get twin registered identity from secrets.
        :param key_seed: twin secrets
        :param key_name: twin key name
        :param twin_did: twin decentralised identifier
        :param twin_name: twin name
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param password: Optional twin password (secrets)
        :return: twin registered identity

        :raises:
            IdentityValidationError: if invalid twin name
        """
        key_pair_secrets = build_twin_secrets(key_seed, key_name, seed_method, password)
        return self._get_identity(key_pair_secrets, twin_did, twin_name)

    def user_delegates_authentication_to_agent(self, user_registered_identity: RegisteredIdentity,
                                               agent_registered_identity: RegisteredIdentity,
                                               delegation_name: str = None):
        """
        User delegates authentication to agent.
        The agent can authenticate on behalf of the user.
        :param user_registered_identity: user registered identity
        :param agent_registered_identity: agent registered identity
        :param delegation_name: Optional delegation name. If None a random name will be chosen

        :raises:
            IdentityValidationError: if registered identities
            IdentityRegisterDocumentKeyConflictError: if authentication delegation proof name is not unique within
                                                      the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
            IdentityDependencyError: if incompatible library dependency
        """
        self.advanced_api.delegate_authentication(user_registered_identity.key_pair_secrets,
                                                  user_registered_identity.issuer.did,
                                                  agent_registered_identity.key_pair_secrets,
                                                  agent_registered_identity.issuer.did,
                                                  delegation_name)

    def twin_delegates_control_to_agent(self, twin_registered_identity: RegisteredIdentity,
                                        agent_registered_identity: RegisteredIdentity,
                                        delegation_name: str = None):
        """
        Twin delegates control to the agent. The agent can control the twin.
        :param twin_registered_identity: twin registered identity
        :param agent_registered_identity: agent registered identity
        :param delegation_name: Optional delegation name. If None a random name will be chosen

        :raises:
           IdentityValidationError: if registered identities
           IdentityRegisterDocumentKeyConflictError: if control delegation proof name is not unique within
                                                     the register document
           IdentityInvalidDocumentError: if invalid register document
           IdentityResolverError: if resolver error
           IdentityDependencyError: if incompatible library dependency
       """
        self.advanced_api.delegate_control(twin_registered_identity.key_pair_secrets,
                                           twin_registered_identity.issuer.did,
                                           agent_registered_identity.key_pair_secrets,
                                           agent_registered_identity.issuer.did,
                                           delegation_name)

    def set_document_controller(self, identity: RegisteredIdentity,
                                controller: Issuer):
        """
        Set controller issuer to the register document associated to the provided registered identity.
        :param identity: registered identity
        :param controller: register document controller issuer

        :raises:
            IdentityValidationError: if invalid registered identity
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(identity.key_pair_secrets)
        self.advanced_api.set_document_controller(key_pair,
                                                  identity.issuer,
                                                  controller)

    def set_document_creator(self, identity: RegisteredIdentity,
                             creator: str):
        """
        Set creator to the register document associated to the provided registered identity.
        :param identity: registered identity
        :param creator: register document creator decentralised identifier

        :raises:
            IdentityValidationError: if invalid registered identity
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(identity.key_pair_secrets)
        self.advanced_api.set_document_creator(key_pair,
                                               identity.issuer,
                                               creator)

    def set_document_revoked(self, identity: RegisteredIdentity,
                             revoked: bool):
        """
        Set register document associated to the provided registered identity revoke field.
        :param identity: registered identity
        :param revoked:  is register document revoked

        :raises:
            IdentityValidationError: if invalid registered identity
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(identity.key_pair_secrets)
        self.advanced_api.set_document_revoked(key_pair,
                                               identity.issuer,
                                               revoked)

    def get_register_document(self, doc_did: str) -> RegisterDocument:
        """
        Get a register document from the resolver.
        :param doc_did: register document decentralised identifier
        :return: associated register document

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverDocNotFoundError: if document not found
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        return self.advanced_api.get_register_document(doc_did)

    def validate_register_document(self, doc: RegisterDocument) -> RegisterDocument:
        """
        Validate a register document against the resolver.
        :param doc: register document

        :raises:
            IdentityInvalidDocumentDelegationError: if one of the register document delegation proof is invalid
        """
        return self.advanced_api.validate_register_document(doc)

    @staticmethod
    def validate_document_proof(doc: RegisterDocument) -> RegisterDocument:
        """
        Verify a register document proof.
        :param doc: register document

        :raises:
            IdentityInvalidDocumentError: if register document proof is invalid
        """
        return AdvancedIdentityLocalApi.validate_document_proof(doc)

    @staticmethod
    def create_agent_auth_token(agent_registered_identity: RegisteredIdentity,
                                user_did: str, duration: int, audience: str = 'default') -> str:
        """
        Create an agent authentication token.
        :param agent_registered_identity: agent registered identity
        :param user_did: user register document decentralised identifier
        :param duration: token duration in seconds
        :param audience: Optional token audience
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if invalid agent registered identity
            IdentityValidationError: if invalid token data
            IdentityDependencyError: if incompatible library dependency
        """
        agent_key_pair = KeyPairSecretsHelper.get_key_pair(agent_registered_identity.key_pair_secrets)
        return AdvancedIdentityLocalApi.create_agent_auth_token(agent_key_pair,
                                                                agent_registered_identity.issuer,
                                                                user_did,
                                                                duration,
                                                                audience)

    @staticmethod
    def get_key_pair_from_user(user_seed: bytes, user_key_name: str,
                               seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                               password: str = '') -> KeyPair:
        """
        Get key pair from user secrets
        :param user_seed: user seed (secrets)
        :param user_key_name: user key name (secrets)
        :param password: Optional user password (secrets)
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :return: user key pair

        :raises:
                IdentityValidationError: if invalid user key seed
                IdentityValidationError: if invalid user key name
        """
        key_pair_secrets = build_user_secrets(user_seed, user_key_name, seed_method, password)
        return KeyPairSecretsHelper.get_key_pair(key_pair_secrets)

    @staticmethod
    def get_key_pair_from_agent(agent_seed: bytes, agent_key_name: str,
                                seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                                password: str = '') -> KeyPair:
        """
        Get key pair from agent secrets
        :param agent_seed: agent seed (secrets)
        :param agent_key_name: agent key name (secrets)
        :param password: Optional agent password (secrets)
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :return: agent key pair

        :raises:
                IdentityValidationError: if invalid agent key seed
                IdentityValidationError: if invalid agent key name
        """
        key_pair_secrets = build_agent_secrets(agent_seed, agent_key_name, seed_method, password)
        return KeyPairSecretsHelper.get_key_pair(key_pair_secrets)

    @staticmethod
    def get_key_pair_from_twin(twin_seed: bytes, twin_key_name: str,
                               seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                               password: str = '') -> KeyPair:
        """
        Get key pair from twin secrets
        :param twin_seed: twin seed (secrets)
        :param twin_key_name: twin key name (secrets)
        :param password: Optional twin password (secrets)
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :return: twin key pair

        :raises:
                IdentityValidationError: if invalid twin key seed
                IdentityValidationError: if invalid twin key name
        """
        key_pair_secrets = build_twin_secrets(twin_seed, twin_key_name, seed_method, password)
        return KeyPairSecretsHelper.get_key_pair(key_pair_secrets)

    def add_new_owner(self, new_owner_name: str, new_owner_public_key_base58: str,
                      doc_registered_identity: RegisteredIdentity) -> Issuer:
        """
        Add new register document owner.
        :param new_owner_name: new owner name
        :param new_owner_public_key_base58: new owner public base 58 key
        :param doc_registered_identity: doc owner registered identity
        :return: new register document owner issuer

        :raises:
            IdentityValidationError: if invalid new owner name
            IdentityValidationError: if invalid doc owner registered identity
            IdentityRegisterDocumentKeyConflictError: if new owner name is not unique within the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(doc_registered_identity.key_pair_secrets)
        return self.advanced_api.add_public_key_to_document(new_owner_name, new_owner_public_key_base58,
                                                            key_pair, doc_registered_identity.issuer)

    def remove_ownership(self, removed_owner_issuer: Issuer,
                         existing_owner_registered_identity: RegisteredIdentity):
        """
        Remove owner from a register document.
        :param removed_owner_issuer: register document owner issuer to remove
        :param existing_owner_registered_identity: other existing doc owner registered identity
        :return: removed register document owner issuer

        :raises:
            IdentityValidationError: if invalid owner
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        existing_owner_key_pair = KeyPairSecretsHelper.get_key_pair(existing_owner_registered_identity.key_pair_secrets)
        return self.advanced_api.remove_public_key_from_document(removed_owner_issuer,
                                                                 existing_owner_key_pair,
                                                                 existing_owner_registered_identity.issuer)

    @staticmethod
    def create_seed(length: Optional[int] = 256) -> bytes:
        """
        Create a new seed (secrets).
        :param length: seed length
        :return: seed

        :raises:
            IdentityValidationError: if invalid seed length
        """
        return AdvancedIdentityLocalApi.create_seed(length)


def get_rest_identity_api(resolver_url, timeout: Optional[Union[int, float]] = None) -> IdentityApi:
    """
    Get a REST identity api.
    :param resolver_url: resolver url
    :param timeout: optional timeout seconds. Default=60s. If set to 0, requests will have no timeout.
    :return: identity api
    """
    resolver_client = get_rest_resolver_client(resolver_url, timeout)
    return IdentityApi(AdvancedIdentityRegisterApi(resolver_client))



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
from dataclasses import dataclass
from typing import Union

from iotics.lib.identity.crypto.proof import Issuer, KeyPairSecrets, Proof
from iotics.lib.identity.register.keys import DelegationProofType


@dataclass(frozen=True)
class APIProof:
    issuer: Issuer
    content: bytes
    signature: str

    @property
    def p_type(self) -> DelegationProofType:
        return DelegationProofType.DID

    @staticmethod
    def build(key_pair: KeyPairSecrets, issuer: Issuer, content: bytes) -> 'APIProof':
        proof = Proof.build(key_pair, issuer, content)
        return APIProof(proof.issuer, proof.content, proof.signature)


@dataclass(frozen=True)
class APIDidDelegationProof(APIProof):

    @property
    def p_type(self) -> DelegationProofType:
        return DelegationProofType.DID

    # pylint: disable=arguments-differ
    @staticmethod
    def build(key_pair: KeyPairSecrets, issuer: Issuer, did: str) -> 'APIDidDelegationProof':
        proof = APIProof.build(key_pair, issuer, did.encode())
        return APIDidDelegationProof(proof.issuer, proof.content, proof.signature)


@dataclass(frozen=True)
class APIGenericDelegationProof(APIProof):

    @property
    def p_type(self) -> DelegationProofType:
        return DelegationProofType.GENERIC

    # pylint: disable=arguments-differ
    @staticmethod
    def build(key_pair: KeyPairSecrets, issuer: Issuer) -> 'APIGenericDelegationProof':
        proof = APIProof.build(key_pair, issuer, b'')
        return APIGenericDelegationProof(proof.issuer, proof.content, proof.signature)


def get_proof_type(proof: Union[Proof, APIProof]) -> DelegationProofType:
    """
    Backward compatibility function to allow legacy Proof usage and new API Delegation Proof usage
    :param proof: a Proof or an API Delegation Proof
    :return: the proof type or the default proof type if not present
    """
    return getattr(proof, 'p_type', DelegationProofType.DID)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
from abc import ABC, abstractmethod
from typing import Optional

from iotics.lib.identity.error import IdentityRegisterDocumentKeyNotFoundError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.document_builder import RegisterDocumentBuilder
from iotics.lib.identity.register.keys import RegisterAuthenticationPublicKey, RegisterDelegationProof, \
    RegisterKeyBase, RegisterPublicKey


class RegisterDocKeyApi(ABC):

    @abstractmethod
    def add_key_to_builder(self, builder: RegisterDocumentBuilder, key: RegisterKeyBase):
        """
        Add a new register key to the register document builder
        :param builder: register document builder
        :param key: register key base
        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        raise NotImplementedError

    @abstractmethod
    def get_key_from_doc(self, doc: RegisterDocument, key_name: str) -> Optional[RegisterKeyBase]:
        """
        Get a register key from the document
        :param doc: register document
        :param key_name: key name
        :return: optional associated register key
        """
        raise NotImplementedError

    def add_doc_key(self, doc: RegisterDocument, key: RegisterKeyBase) -> RegisterDocument:
        """
        Add a new register key to the register document
        :param doc: register document
        :param key: register key base

        :raises:
            IdentityInvalidDocumentError: if invalid document
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        builder = RegisterDocumentBuilder()
        self.add_key_to_builder(builder, key)
        return builder.build_from_existing(doc)

    @staticmethod
    def remove_doc_key(doc: RegisterDocument, key_name: str) -> RegisterDocument:
        """
        Remove a register key from a register document
        :param doc: register document
        :param key_name: register key name
        :return:
        """
        return RegisterDocumentBuilder() \
            .set_keys_from_existing(doc) \
            .remove_key(key_name) \
            .build_from_existing(doc, populate_with_doc_keys=False)

    def revoke_doc_key(self, doc: RegisterDocument, key_name: str, revoked: bool) -> RegisterDocument:
        """
        Create a new document setting revoked to the key associated to the key name
        :param doc: a register document
        :param key_name: a key name
        :param revoked: is the key revoked
        :return: a register document

        :raises:
        - IdentityRegisterDocumentKeyNotFoundError: if the key to revoke is not found

        """
        key = self.get_key_from_doc(doc, key_name)
        if not key:
            raise IdentityRegisterDocumentKeyNotFoundError(f'Can mot revoke key {key_name} fron document {doc.did}:'
                                                           f'key not found')
        builder = RegisterDocumentBuilder() \
            .set_keys_from_existing(doc) \
            .remove_key(key_name)
        self.add_key_to_builder(builder, key.get_new_key(revoked))
        return builder.build_from_existing(doc, populate_with_doc_keys=False)


class RegisterPublicDocKeysApi(RegisterDocKeyApi):

    def get_key_from_doc(self, doc: RegisterDocument, key_name: str) -> Optional[RegisterKeyBase]:
        return doc.public_keys.get(key_name)

    def add_key_to_builder(self, builder: RegisterDocumentBuilder, key: RegisterPublicKey):  # type: ignore
        builder.add_public_key_obj(key)


class RegisterAuthPublicDocKeysApi(RegisterDocKeyApi):
    def get_key_from_doc(self, doc: RegisterDocument, key_name: str) -> Optional[RegisterKeyBase]:
        return doc.auth_keys.get(key_name)

    def add_key_to_builder(self, builder: RegisterDocumentBuilder,  # type: ignore
                           key: RegisterAuthenticationPublicKey):
        builder.add_authentication_key_obj(key)


class RegisterCtrlDelegPublicDocKeysApi(RegisterDocKeyApi):

    def get_key_from_doc(self, doc: RegisterDocument, key_name: str) -> Optional[RegisterKeyBase]:
        return doc.control_delegation_proof.get(key_name)

    def add_key_to_builder(self, builder: RegisterDocumentBuilder, key: RegisterDelegationProof):  # type: ignore
        builder.add_control_delegation_obj(key)


class RegisterAuthDelegPublicDocKeysApi(RegisterDocKeyApi):

    def get_key_from_doc(self, doc: RegisterDocument, key_name: str) -> Optional[RegisterKeyBase]:
        return doc.auth_delegation_proof.get(key_name)

    def add_key_to_builder(self, builder: RegisterDocumentBuilder, key: RegisterDelegationProof):  # type: ignore
        builder.add_authentication_delegation_obj(key)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

# Secrets validation

MIN_SEED_METHOD_NONE_LEN = 16
KEY_PAIR_PATH_PREFIX = 'iotics/0'

# Identifier Validation
NAME_PATTERN_RAW = r'[a-zA-Z\-\_0-9]{1,24}'
ISSUER_SEPARATOR = '#'
IDENTIFIER_PREFIX = 'did:iotics:'
IDENTIFIER_ID = fr'{IDENTIFIER_PREFIX}iot(?P<hash>[a-km-zA-HJ-NP-Z1-9]{{33}})'
IDENTIFIER_ID_PATTERN = fr'^{IDENTIFIER_ID}$'
IDENTIFIER_NAME_PATTERN = fr'^\{ISSUER_SEPARATOR}{NAME_PATTERN_RAW}$'
ISSUER_PATTERN = fr'^{IDENTIFIER_ID}\#{NAME_PATTERN_RAW}'

# Document constants
DOCUMENT_CONTEXT = 'https://w3id.org/did/v1'
DOCUMENT_VERSION = '0.0.1'
SUPPORTED_VERSIONS = ('0.0.0', DOCUMENT_VERSION)
# public key type string
DOCUMENT_PUBLIC_KEY_TYPE = 'Secp256k1VerificationKey2018'
# authentication public key type string
DOCUMENT_AUTHENTICATION_TYPE = 'Secp256k1SignatureAuthentication2018'
# Document metadata validation
DOCUMENT_MAX_LABEL_LENGTH = 64
DOCUMENT_MAX_COMMENT_LENGTH = 512
DOCUMENT_MAX_URL_LENGTH = 512

# Token constants
# Default offset for token valid-from time used. This is to avoid tokens being rejected when the client time is
# marginally ahead of the server (i.e. resolver or Iotics host).
DEFAULT_TOKEN_START_OFFSET_SECONDS = -30
TOKEN_ALGORITHM = 'ES256'



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

import re

from iotics.lib.identity.const import IDENTIFIER_ID_PATTERN, IDENTIFIER_NAME_PATTERN, ISSUER_PATTERN
from iotics.lib.identity.error import IdentityValidationError


class IdentityValidation:
    @staticmethod
    def validate_identifier(did: str):
        """
        Validate decentralised identifier.
        :param did: decentralised identifier

        :raises:
            IdentityValidationError: if invalid identifier
        """
        result = re.match(IDENTIFIER_ID_PATTERN, did)
        if result is None:
            raise IdentityValidationError(f'Identifier does not match pattern {did} - {IDENTIFIER_ID_PATTERN}')

    @staticmethod
    def validate_issuer_string(issuer: str):
        """
        Validate issuer.
        :param issuer: issuer as string

        :raises:
            IdentityValidationError: if invalid issuer
        """
        result = re.match(ISSUER_PATTERN, issuer)
        if result is None:
            raise IdentityValidationError(f'Identifier does not match pattern {issuer} - {ISSUER_PATTERN}')

    @staticmethod
    def validate_key_name(name: str) -> bool:
        """
        Validate key name.
        :param name: key name

        :raises:
            IdentityValidationError: if invalid key name
        """
        m = re.match(IDENTIFIER_NAME_PATTERN, name)
        if m is None:
            raise IdentityValidationError(f'Name is not valid: {name} - {IDENTIFIER_NAME_PATTERN}')
        return True



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

import random
import string
from typing import Callable, Optional

import base58

from iotics.lib.identity.const import ISSUER_SEPARATOR
from iotics.lib.identity.crypto.identity import make_identifier
from iotics.lib.identity.crypto.issuer import Issuer, IssuerKey
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.keys import RegisterDelegationProof, RegisterKey, RegisterPublicKey
from iotics.lib.identity.validation.identity import IdentityValidation

GetControllerDocFunc = Callable[[str], RegisterDocument]


class RegisterDocumentHelper:
    @staticmethod
    def get_owner_register_public_key(doc: RegisterDocument) -> Optional[RegisterPublicKey]:
        """
        Get the register document initial owner public key
        :param doc: existing register document
        :return: RegisterPublicKey if found or None
        """
        for key in doc.public_keys.values():
            public_bytes = base58.b58decode(key.base58)
            key_id = make_identifier(public_bytes)
            if key_id == doc.did:  # It is the original key
                return key
        return None

    @staticmethod
    def get_issuer_from_public_key(doc: RegisterDocument, public_base58: str) -> Optional[Issuer]:
        """
        Find a register key by issuer and returns True/False if found.
        Lookup in the register document public keys (and authentication keys if include_auth is set to True).
        :param doc: existing register document
        :param public_base58: public key to search for as base58 string
        :return: Issuer or None

        :raises:
            IdentityValidationError: if invalid name or did
        """
        for key in doc.public_keys.values():
            if key.base58 == public_base58:
                return Issuer.build(doc.did, key.name)
        return None

    @staticmethod
    def is_issuer_in_keys(issuer_name: str, doc: RegisterDocument, include_auth: bool) -> bool:
        """
        Find an issuer key and returns True/False if found
        :param issuer_name: #name to search for
        :param doc: existing register document
        :param include_auth: include authentication keys
        :return: True/False if found
        """
        return RegisterDocumentHelper.get_issuer_register_key(issuer_name, doc, include_auth) is not None

    @staticmethod
    def get_issuer_register_key(issuer_name: str, doc: RegisterDocument, include_auth: bool) -> Optional[RegisterKey]:
        """
        Find a register key by issuer.
        Lookup in the register document public keys (and authentication keys if include_auth is set to True).
        :param issuer_name: #name to search for
        :param doc: existing register document
        :param include_auth: include authentication keys
        :return: RegisterKey or None
        """
        pub_key = doc.public_keys.get(issuer_name)
        if pub_key:
            return pub_key
        if include_auth:
            return doc.auth_keys.get(issuer_name)
        return None

    @staticmethod
    def get_issuer_register_delegation_proof(issuer_name: str, doc: RegisterDocument,
                                             include_auth: bool) -> Optional[RegisterDelegationProof]:
        """
        Find a register delegation proof by issuer.
        Lookup in the register document control delegation proofs
        (and authentication delegation proofs if include_auth is set to True).
        :param issuer_name: #name to search for
        :param doc: existing register document
        :param include_auth: include authentication keys
        :return: RegisterDelegationProof or None
        """
        control_deleg = doc.control_delegation_proof.get(issuer_name)
        if control_deleg:
            return control_deleg
        if include_auth:
            return doc.auth_delegation_proof.get(issuer_name)
        return None

    @staticmethod
    def get_register_delegation_proof_by_controller(controller: Issuer, doc: RegisterDocument,
                                                    include_auth: bool) -> Optional[RegisterDelegationProof]:
        """
        Find a register delegation proof by controller issuer.
        Lookup in the register document control delegation proofs
        (and authentication delegation proofs if include_auth is set to True).
        :param controller: Issuing controller
        :param doc: existing register document
        :param include_auth: include authentication keys
        :return: RegisterDelegationProof or None
        """
        keys = list(doc.control_delegation_proof.values())
        if include_auth:
            keys += list(doc.auth_delegation_proof.values())
        for key in keys:
            if key.controller == controller:
                return key
        return None

    @staticmethod
    def get_valid_issuer_key_for(doc: RegisterDocument, issuer_name: str, get_controller_doc: GetControllerDocFunc,
                                 include_auth: bool, ) -> Optional[IssuerKey]:
        """
        Get a valid issuer key matching issuer name
        :param doc: existing register document
        :param issuer_name: name of issuer
        :param get_controller_doc: resolver discover function
        :param include_auth: include authentication keys
        :return: IssuerKey or None

        :raises:
            IdentityValidationError: if invalid name or did
        """
        key = RegisterDocumentHelper.get_issuer_register_key(issuer_name, doc, include_auth)
        if key:
            return IssuerKey.build(doc.did, issuer_name, key.base58)

        deleg_key = RegisterDocumentHelper.get_issuer_register_delegation_proof(issuer_name, doc, include_auth)
        if deleg_key:
            controlled_doc = get_controller_doc(deleg_key.controller.did)
            key = RegisterDocumentHelper.get_issuer_register_key(issuer_name, controlled_doc, include_auth)
            if key:
                return IssuerKey.build(doc.did, issuer_name, key.base58)
        return None

    @staticmethod
    def get_valid_issuer_key_for_control_only(doc: RegisterDocument, issuer_name: str,
                                              get_controller_doc: GetControllerDocFunc) -> Optional[IssuerKey]:
        """
        Get a valid issuer key matching issuer name from the control keys and delegation proofs only
        :param doc: existing register document
        :param issuer_name: name of issuer
        :param get_controller_doc: resolver discover function
        :return: IssuerKey or None

        :raises:
            IdentityValidationError: if invalid name or did
        """
        return RegisterDocumentHelper.get_valid_issuer_key_for(doc, issuer_name, get_controller_doc, include_auth=False)

    @staticmethod
    def get_valid_issuer_key_for_auth(doc: RegisterDocument, issuer_name: str,
                                      get_controller_doc: GetControllerDocFunc) -> Optional[IssuerKey]:
        """
        Get a valid issuer key matching issuer name from the control and authentication keys and delegation proofs
        :param doc: existing register document
        :param issuer_name: name of issuer
        :param get_controller_doc: resolver discover function
        :return: IssuerKey or None

        :raises:
            IdentityValidationError: if invalid name or did
        """
        return RegisterDocumentHelper.get_valid_issuer_key_for(doc, issuer_name, get_controller_doc, include_auth=True)

    @staticmethod
    def new_random_name_for_document(doc: RegisterDocument, length: int = 10) -> str:
        """
        New random #name that can be added to this document
        :param doc: existing register document
        :param length: length of name including #
        :return: a new random #name (not already existing in this document)

        ;raises:
            IdentityValidationError: if invalid key name (eg if length too short/long)
        """
        length -= 1
        chars = string.ascii_lowercase + string.digits
        while True:
            new_name = ISSUER_SEPARATOR + ''.join(random.choice(chars) for _ in range(length))
            IdentityValidation.validate_key_name(new_name)

            if not RegisterDocumentHelper.get_issuer_register_key(new_name, doc, True) and \
               not RegisterDocumentHelper.get_issuer_register_delegation_proof(new_name, doc, True):
                return new_name



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

import base64

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.crypto.identity import is_same_identifier
from iotics.lib.identity.crypto.keys import KeysHelper
from iotics.lib.identity.crypto.proof import Proof
from iotics.lib.identity.error import IdentityInvalidDocumentDelegationError, IdentityInvalidProofError, \
    IdentityResolverError, \
    IdentityValidationError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.keys import DelegationProofType, RegisterDelegationProof, RegisterPublicKey
from iotics.lib.identity.register.resolver import ResolverClient


class ProofValidation:
    @staticmethod
    def validate_proof(proof: Proof, public_base58: str):
        """
        Validate proof.
        :param proof: proof
        :param public_base58: public key base 58 used to create the proof

        :raises:
            IdentityInvalidProofError: if invalid proof signature
            IdentityInvalidProofError: if invalid proof
        """
        public_ecdsa = KeysHelper.get_public_ECDSA_from_base58(public_base58)
        try:
            signature = base64.b64decode(proof.signature)
            public_ecdsa.verify(signature, proof.content, ec.ECDSA(hashes.SHA256()))
        except InvalidSignature as err:
            raise IdentityInvalidProofError('Invalid proof: invalid signature') from err
        except ValueError as err:
            raise IdentityInvalidProofError(f'Invalid proof: \'{err}\'') from err


class DelegationValidation:
    @staticmethod
    def _is_valid_issuer_or_reusable_proof(deleg_proof: RegisterDelegationProof, public_key: RegisterPublicKey,
                                           doc_id: str):
        controller_issuer = deleg_proof.controller
        if deleg_proof.proof_type == DelegationProofType.DID:
            proof = Proof(controller_issuer, doc_id.encode('ascii'), deleg_proof.proof)
            ProofValidation.validate_proof(proof, public_key.base58)
        elif deleg_proof.proof_type == DelegationProofType.GENERIC:
            proof = Proof(controller_issuer, b'', deleg_proof.proof)
            ProofValidation.validate_proof(proof, public_key.base58)
        else:
            raise IdentityInvalidProofError(f'Invalid proof: invalid type {deleg_proof.proof_type}')

    @staticmethod
    def validate_delegation_from_doc(doc_id: str, controller_doc: RegisterDocument,
                                     deleg_proof: RegisterDelegationProof):
        """
        Validate register delegation proof against the deleagtion controller register document.
        :param doc_id: decentralised id of the register document owning the register delegation proof
        :param controller_doc: delegation controller register document
        :param deleg_proof: register delegation proof under validation

        :raises:
            IdentityInvalidDocumentDelegationError: if controller issuer does not belongs to the controller document
                                                    public keys
            IdentityInvalidDocumentDelegationError: if invalid register delegation proof signature
        """
        try:
            controller_issuer = deleg_proof.controller
            public_key = controller_doc.public_keys.get(controller_issuer.name)
            if not public_key:
                raise IdentityValidationError(f'Public key \'{controller_issuer.name}\' not found'
                                              f' on controller doc \'{controller_doc.did}\'')
            DelegationValidation._is_valid_issuer_or_reusable_proof(deleg_proof, public_key, doc_id)
        except IdentityValidationError as err:
            raise IdentityInvalidDocumentDelegationError(f'Invalid delegation for doc \'{doc_id}\''
                                                         f' with controller: \'{deleg_proof.name}\': {err}') from err

    @staticmethod
    def validate_delegation(resolver_client: ResolverClient, doc_id: str, deleg_proof: RegisterDelegationProof):
        """
        Validate register delegation proof.
        :param resolver_client: resolver client interface
        :param doc_id: decentralised id of the register document owning the register delegation proof
        :param deleg_proof: register delegation proof under validation

        :raises:
            IdentityInvalidDocumentDelegationError: if the register delegation proof is invalid
            IdentityInvalidDocumentDelegationError: if the register delegation proof controller can not be fetched
                                                    from the resolver
        """
        try:
            if is_same_identifier(doc_id, str(deleg_proof.controller)):
                raise IdentityInvalidDocumentDelegationError(f'Delegation on self no allowed on doc \'{doc_id}\'')

            controller_doc = resolver_client.get_document(deleg_proof.controller.did)
            DelegationValidation.validate_delegation_from_doc(doc_id, controller_doc, deleg_proof)
        except IdentityResolverError as err:
            raise IdentityInvalidDocumentDelegationError(f'Invalid delegation for doc \'{doc_id}\''
                                                         f' with controller: \'{deleg_proof.name}\': {err}') from err



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from dataclasses import dataclass

from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.key_pair_secrets import KeyPairSecrets


@dataclass(frozen=True)
class RegisteredIdentity:
    key_pair_secrets: KeyPairSecrets
    issuer: Issuer

    @property
    def did(self):
        return self.issuer.did

    @property
    def name(self):
        return self.issuer.name



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

import hmac
from dataclasses import dataclass
from enum import Enum, unique
from hashlib import sha256, sha512

from cryptography.hazmat.primitives.asymmetric import ec
from mnemonic import Mnemonic  # type: ignore
from mnemonic.mnemonic import ConfigurationError  # type: ignore

from iotics.lib.identity.const import KEY_PAIR_PATH_PREFIX, MIN_SEED_METHOD_NONE_LEN
from iotics.lib.identity.crypto.keys import KeyPair, KeysHelper
from iotics.lib.identity.error import IdentityDependencyError, IdentityValidationError


@unique
class DIDType(Enum):
    HOST = 'host'  # Not used in Iotics ecosystem
    USER = 'user'
    AGENT = 'agent'
    TWIN = 'twin'

    def __str__(self) -> str:
        return str(self.value)


@unique
class SeedMethod(Enum):
    SEED_METHOD_NONE = 0
    SEED_METHOD_BIP39 = 1


@dataclass(frozen=True)
class KeyPairSecrets:
    seed: bytes
    path: str
    seed_method: SeedMethod
    password: str

    @staticmethod
    def build(seed: bytes, path: str, seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
              password: str = '') -> 'KeyPairSecrets':
        """
        Build a valid key pair secrets.
        :param seed: key seed
        :param path: key path
        :param seed_method: seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
        :param password: key password (secrets) (Optional: can be empty string)
        :return: valid key pair secrets

        :raises:
            IdentityValidationError: if invalid key seed
            IdentityValidationError: if invalid key path
        """
        if seed_method == SeedMethod.SEED_METHOD_NONE:
            if len(seed) < MIN_SEED_METHOD_NONE_LEN:
                raise IdentityValidationError(f'Invalid seed length for method \'SEED_METHOD_NONE\', '
                                              f'must be at least {MIN_SEED_METHOD_NONE_LEN} bytes')
        elif seed_method == SeedMethod.SEED_METHOD_BIP39:
            KeyPairSecretsHelper.validate_bip39_seed(seed)
        else:
            raise IdentityValidationError(f'Invalid seed method \'{seed_method}\', '
                                          f'must be in {[m.name for m in SeedMethod]}')
        if not path.startswith(KEY_PAIR_PATH_PREFIX):
            raise IdentityValidationError(f'Invalid key pair path \'{path}\', '
                                          f'must start with {KEY_PAIR_PATH_PREFIX}')

        return KeyPairSecrets(seed, path, seed_method, password)


def build_user_secrets(seed: bytes, name: str,
                       seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                       password: str = '') -> KeyPairSecrets:
    """
    Build a valid key pair secrets for User.
    :param seed: User key seed
    :param name: User key name
    :param seed_method: User seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
    :param password: User key password (secrets) (Optional: can be empty string)
    :return: valid User key pair secrets

    :raises:
        IdentityValidationError: if invalid User key seed
        IdentityValidationError: if invalid User key name
    """
    user_path = f'{KEY_PAIR_PATH_PREFIX}/{DIDType.USER}/{name}'
    return KeyPairSecrets.build(seed, user_path, seed_method, password)


def build_agent_secrets(seed: bytes, name: str,
                        seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                        password: str = '') -> KeyPairSecrets:
    """
    Build a valid key pair secrets for Agent.
    :param seed: Agent key seed
    :param name: Agent key name
    :param seed_method: Agent seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
    :param password: Agent key password (secrets) (Optional: can be empty string)
    :return: valid Agent key pair secrets

    :raises:
        IdentityValidationError: if invalid Agent key seed
        IdentityValidationError: if invalid Agent key name
    """
    agent_path = f'{KEY_PAIR_PATH_PREFIX}/{DIDType.AGENT}/{name}'
    return KeyPairSecrets.build(seed, agent_path, seed_method, password)


def build_twin_secrets(seed: bytes, name: str,
                       seed_method: SeedMethod = SeedMethod.SEED_METHOD_BIP39,
                       password: str = '') -> KeyPairSecrets:
    """
    Build a valid key pair secrets for Twin.
    :param seed: Twin key seed
    :param name: Twin key name
    :param seed_method: Twin seed method (SEED_METHOD_NONE or SEED_METHOD_BIP39) default=SEED_METHOD_BIP39
    :param password: Twin key password (secrets) (Optional: can be empty string)
    :return: valid Twin key pair secrets

    :raises:
        IdentityValidationError: if invalid Twin key seed
        IdentityValidationError: if invalid Twin key name
    """
    twin_path = f'{KEY_PAIR_PATH_PREFIX}/{DIDType.TWIN}/{name}'
    return KeyPairSecrets.build(seed, twin_path, seed_method, password)


class KeyPairSecretsHelper:

    @staticmethod
    def mnemonic_bip39_to_seed(mnemonic: str, lang: str = 'english') -> bytes:
        """
        Take mnemonic string and return seed bytes
        :param mnemonic: mnemonic string
        :param lang: language
        :return: seed bytes

        :raises:
            IdentityValidationError: if invalid seed
            IdentityValidationError: if invalid lang
            IdentityDependencyError: if incompatible Mnemonic dependency
        """
        try:
            men = Mnemonic(lang)
            return men.to_entropy(mnemonic)
        except ConfigurationError as err:
            raise IdentityDependencyError(f'Dependency Mnemonic Internal Error: {err}') from err
        except (LookupError, ValueError) as err:
            raise IdentityValidationError(f'{err}') from err
        except OSError as err:
            raise IdentityValidationError(f'Invalid language for mnemonic: {err}') from err

    @staticmethod
    def seed_bip39_to_mnemonic(seed: bytes, lang: str = 'english') -> str:
        """
        Convert a BIP39 seed to mnemonic.
        :param seed: BIP39 seed
        :param lang: a mnemonic language
        :return: a mnemonic

        :raises:
            IdentityValidationError: if invalid seed
            IdentityValidationError: if invalid lang
            IdentityDependencyError: if incompatible Mnemonic dependency
        """
        try:
            men = Mnemonic(lang)
            return men.to_mnemonic(seed)
        except ConfigurationError as err:
            raise IdentityDependencyError(f'Dependency Mnemonic Internal Error: {err}') from err
        except TypeError as err:
            raise IdentityValidationError(f'Invalid seed format for method \'SEED_METHOD_BIP39_LEN\': {err}') from err
        except ValueError as err:
            raise IdentityValidationError(f'Invalid seed length for method \'SEED_METHOD_BIP39_LEN\': {err}') from err
        except OSError as err:
            raise IdentityValidationError(f'Invalid language for mnemonic: {err}') from err

    @staticmethod
    def validate_bip39_seed(seed: bytes):
        """
        Valid BIP39 seed
        :param seed: BIP39 seed

        :raises:
            IdentityValidationError: if invalid seed
            IdentityValidationError: if invalid lang
            IdentityDependencyError: if incompatible Mnemonic dependency

        """
        KeyPairSecretsHelper.seed_bip39_to_mnemonic(seed)

    @staticmethod
    def get_private_key(key_pair_secrets: KeyPairSecrets) -> ec.EllipticCurvePrivateKey:
        """
        Get private key from key pair secrets
        :param key_pair_secrets: key pair secrets
        :return: private key

        :raise:
            IdentityValidationError: if invalid seed method
            IdentityValidationError: if invalid lang
            IdentityDependencyError: if incompatible Mnemonic dependency
            IdentityDependencyError: if incompatible EllipticCurve dependency
        """
        if key_pair_secrets.seed_method == SeedMethod.SEED_METHOD_NONE:
            result = hmac.new(key_pair_secrets.seed, key_pair_secrets.password.encode(), sha512).digest()
        elif key_pair_secrets.seed_method == SeedMethod.SEED_METHOD_BIP39:
            men = KeyPairSecretsHelper.seed_bip39_to_mnemonic(key_pair_secrets.seed)
            result = Mnemonic.to_seed(men, key_pair_secrets.password)
        else:
            raise IdentityValidationError(f'Invalid seed method \'{key_pair_secrets.seed_method}\', '
                                          f'must be in {[m.name for m in SeedMethod]}')

        private_expo = hmac.new(result, key_pair_secrets.path.encode(), sha256).hexdigest()
        return KeysHelper.get_private_ECDSA(private_expo)

    @staticmethod
    def get_public_key_base58_from_key_pair_secrets(key_pair_secrets: KeyPairSecrets) -> str:
        """
        Get public key base58 fron key pair secrets
        :param key_pair_secrets: key pair secrets
        :return: public key base58

        :raise:
            IdentityValidationError: if invalid key_pair_secrets
            IdentityDependencyError: if incompatible EllipticCurve dependency
        """
        private_key = KeyPairSecretsHelper.get_private_key(key_pair_secrets)
        _, base58 = KeysHelper.get_public_keys_from_private_ECDSA(private_key)
        return base58

    @staticmethod
    def get_key_pair(key_pair_secrets: KeyPairSecrets) -> KeyPair:
        """
        Get key pair from key pair secrets
        :param key_pair_secrets: key pair secrets
        :return: key pair

        :raises:
            IdentityValidationError: if invalid key_pair_secrets
        """
        private_key = KeyPairSecretsHelper.get_private_key(key_pair_secrets)
        public_bytes, public_base58 = KeysHelper.get_public_keys_from_private_ECDSA(private_key)
        return KeyPair(private_key=private_key,
                       public_bytes=public_bytes,
                       public_base58=public_base58)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from dataclasses import dataclass
from typing import Dict, Optional

from iotics.lib.identity.const import DOCUMENT_CONTEXT, DOCUMENT_MAX_COMMENT_LENGTH, DOCUMENT_MAX_LABEL_LENGTH, \
    DOCUMENT_MAX_URL_LENGTH, DOCUMENT_VERSION
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.key_pair_secrets import DIDType
from iotics.lib.identity.error import IdentityValidationError
from iotics.lib.identity.register.keys import RegisterAuthenticationPublicKey, RegisterDelegationProof, \
    RegisterPublicKey


@dataclass(frozen=True)
class Metadata:
    label: Optional[str] = None
    comment: Optional[str] = None
    url: Optional[str] = None

    def to_dict(self):
        ret = {}
        if self.label:
            ret['label'] = self.label
        if self.comment:
            ret['comment'] = self.comment
        if self.url:
            ret['url'] = self.url
        return ret

    @staticmethod
    def from_dict(data: dict):
        """
        Build register metadata from dict.
        :param data: register metadata as dict
        :return: valid register metadata

        :raises:
            IdentityValidationError: if invalid metadata as dict
        """
        return Metadata.build(data.get('label'), data.get('comment'), data.get('url'))

    @staticmethod
    def build(label: Optional[str], comment: Optional[str], url: Optional[str]):
        """
        Build register metadata.
        :param label: metadata label
        :param comment: metadata comment
        :param url: metadata url
        :return: valid register metadata

        :raises:
            IdentityValidationError: if invalid label
            IdentityValidationError: if invalid comment
            IdentityValidationError: if invalid url
        """
        if label and len(label) > DOCUMENT_MAX_LABEL_LENGTH:
            raise IdentityValidationError(f'Document metadata label it too long, max size: \'{label}\'')
        if comment and len(comment) > DOCUMENT_MAX_COMMENT_LENGTH:
            raise IdentityValidationError(f'Document metadata comment it too long, max size: \'{comment}\'')
        if url and len(url) > DOCUMENT_MAX_URL_LENGTH:
            raise IdentityValidationError(f'Document metadata url it too long, max size: \'{url}\'')
        return Metadata(label, comment, url)


@dataclass(frozen=True)
class RegisterDocument:
    public_keys: Dict[str, RegisterPublicKey]
    auth_keys: Dict[str, RegisterAuthenticationPublicKey]
    auth_delegation_proof: Dict[str, RegisterDelegationProof]
    control_delegation_proof: Dict[str, RegisterDelegationProof]
    did: str
    purpose: DIDType
    proof: str
    revoked: bool
    spec_version: str = DOCUMENT_VERSION
    metadata: Metadata = Metadata()
    creator: Optional[str] = None
    update_time: Optional[int] = None
    controller: Optional[Issuer] = None

    def to_dict(self) -> dict:
        """
        Serialise thee register document to dict.
        :return: register document as dict
        """
        ret = {
            '@context': DOCUMENT_CONTEXT,
            'id': self.did,
            'ioticsSpecVersion': self.spec_version,
            'ioticsDIDType': self.purpose.value,
            'updateTime': self.update_time,
            'proof': self.proof,
            'publicKey': [k.to_dict() for _, k in self.public_keys.items()],
            'authentication': [k.to_dict() for _, k in self.auth_keys.items()],
            'delegateControl': [k.to_dict() for _, k in self.control_delegation_proof.items()],
            'delegateAuthentication': [k.to_dict() for _, k in self.auth_delegation_proof.items()],
            'metadata': self.metadata.to_dict(),
            'revoked': self.revoked

        }
        if self.controller:
            ret['controller'] = str(self.controller)
        if self.creator:
            ret['creator'] = self.creator
        return ret



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
import secrets
from typing import Callable, Optional, Tuple

from iotics.lib.identity.api.advanced_api_keys import RegisterAuthDelegPublicDocKeysApi, RegisterAuthPublicDocKeysApi, \
    RegisterCtrlDelegPublicDocKeysApi, RegisterPublicDocKeysApi
from iotics.lib.identity.api.advanced_api_proof import APIProof, APIDidDelegationProof, \
    APIGenericDelegationProof, get_proof_type
from iotics.lib.identity.const import DEFAULT_TOKEN_START_OFFSET_SECONDS, ISSUER_SEPARATOR
from iotics.lib.identity.crypto.identity import make_identifier
from iotics.lib.identity.crypto.issuer import Issuer, IssuerKey
from iotics.lib.identity.crypto.jwt import JwtTokenHelper
from iotics.lib.identity.crypto.key_pair_secrets import DIDType, KeyPairSecrets, KeyPairSecretsHelper
from iotics.lib.identity.crypto.keys import KeyPair, KeysHelper
from iotics.lib.identity.crypto.proof import Proof
from iotics.lib.identity.error import IdentityRegisterIssuerNotFoundError, IdentityResolverConflictError, \
    IdentityResolverDocNotFoundError, IdentityValidationError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.document_builder import RegisterDocumentBuilder
from iotics.lib.identity.register.document_helper import RegisterDocumentHelper
from iotics.lib.identity.register.key_pair import RegisteredIdentity
from iotics.lib.identity.register.keys import DelegationProofType, RegisterDelegationProof, RegisterPublicKey
from iotics.lib.identity.register.resolver import ResolverClient
from iotics.lib.identity.validation.document import DocumentValidation


class AdvancedIdentityLocalApi:
    @staticmethod
    def get_key_pair_from_hex_private_key(private_expo_hex: str) -> KeyPair:
        """
        Get keypair given the private exponent as a hex string
        :param private_expo_hex: private exponent as hex string
        :return: KeyPair instance

        :raises:
            IdentityDependencyError: if incompatible EllipticCurve dependency
        """
        private_key = KeysHelper.get_private_ECDSA(private_expo_hex)
        public_bytes, public_base58 = KeysHelper.get_public_keys_from_private_ECDSA(private_key)
        return KeyPair(private_key=private_key,
                       public_bytes=public_bytes,
                       public_base58=public_base58)

    @staticmethod
    def get_issuer_by_public_key(doc: RegisterDocument, public_base58: str) -> Issuer:
        """
        Get issuer matching the public key from a register document public keys.
        :param doc: register document
        :param public_base58: issuer public key base 58
        :return: corresponding issuer

        :raises:
            IdentityRegisterIssuerNotFoundError: if the issuer matching the public key is not found
        """
        issuer = RegisterDocumentHelper.get_issuer_from_public_key(doc, public_base58)
        if not issuer:
            raise IdentityRegisterIssuerNotFoundError(f'User secrets not allowed to update \'{doc.did}\'')
        return issuer

    @staticmethod
    def create_agent_auth_token(agent_key_pair: KeyPair, agent_issuer: Issuer, user_did: str, duration: int,
                                audience: str, start_offset: int = DEFAULT_TOKEN_START_OFFSET_SECONDS):
        """
        Create an agent authentication token.
        :param agent_key_pair: agent key pair
        :param agent_issuer: agent issuer
        :param user_did: user register document decentralised identifier
        :param duration: token duration in seconds
        :param audience: Optional token audience
        :param start_offset: Optional offset for token valid-from time used (default=DEFAULT_TOKEN_START_OFFSET_SECONDS)
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if invalid agent secrets
            IdentityValidationError: if invalid token data
            IdentityDependencyError: if incompatible library dependency
        """
        return JwtTokenHelper.create_auth_token(str(agent_issuer),
                                                user_did,
                                                audience,
                                                duration,
                                                agent_key_pair.private_key,
                                                start_offset)

    @staticmethod
    def create_twin_token(twin_key_pair: KeyPair, twin_issuer: Issuer, duration: int,
                          audience: str, start_offset: int = DEFAULT_TOKEN_START_OFFSET_SECONDS):
        """
        Create a twin authentication token.
        :param twin_key_pair: twin key pair
        :param twin_issuer: twin issuer
        :param duration: token duration in seconds
        :param audience: Optional token audience
        :param start_offset: Optional offset for token valid-from time used (default=DEFAULT_TOKEN_START_OFFSET_SECONDS)
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if invalid twin secrets
            IdentityValidationError: if invalid token data
            IdentityDependencyError: if incompatible library dependency
        """
        return JwtTokenHelper.create_auth_token(str(twin_issuer),
                                                twin_issuer.did,
                                                audience,  # type: ignore
                                                duration,
                                                twin_key_pair.private_key,
                                                start_offset)

    @staticmethod
    def create_proof(key_secrets: KeyPairSecrets, issuer: Issuer, content: bytes) -> APIProof:
        """
        Create a proof.
        :param key_secrets: secrets used to build the proof signature
        :param issuer: proof issuer
        :param content: proof content
        :return: proof

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityDependencyError: if incompatible library dependency
        """
        return APIProof.build(key_secrets, issuer, content=content)

    @staticmethod
    def create_delegation_proof(delegating_issuer: Issuer,
                                subject_doc: RegisterDocument,
                                subject_secrets: KeyPairSecrets) -> Tuple[Issuer, APIDidDelegationProof]:
        """
        Create a delegation proof that can be used to setup a delegation from a single delegating issuer.
        :param delegating_issuer: delegating issuer
        :param subject_doc: subject/delegated register document
        :param subject_secrets: subject/delegated secrets
        :return: subject/delegated issuer and proof

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityRegisterIssuerNotFoundError: if the issuer matching the subject/delegated secrets is not found
            IdentityDependencyError: if incompatible library dependency
        """
        subject_key_pair = KeyPairSecretsHelper.get_key_pair(subject_secrets)
        subject_issuer = AdvancedIdentityLocalApi.get_issuer_by_public_key(subject_doc,
                                                                           subject_key_pair.public_base58)
        proof = APIDidDelegationProof.build(subject_secrets, subject_issuer, delegating_issuer.did)
        return subject_issuer, proof

    @staticmethod
    def create_generic_delegation_proof(subject_doc: RegisterDocument,
                                        subject_secrets: KeyPairSecrets) -> Tuple[Issuer, APIGenericDelegationProof]:
        """
        Create a delegation proof that can be re-used to setup a delegation from several delegating issuers.
        :param subject_doc: subject/delegated register document
        :param subject_secrets: subject/delegated secrets
        :return: subject/delegated issuer and proof

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityRegisterIssuerNotFoundError: if the issuer matching the subject/delegated secrets is not found
            IdentityDependencyError: if incompatible library dependency
        """
        subject_key_pair = KeyPairSecretsHelper.get_key_pair(subject_secrets)
        subject_issuer = AdvancedIdentityLocalApi.get_issuer_by_public_key(subject_doc,
                                                                           subject_key_pair.public_base58)
        proof = APIGenericDelegationProof.build(subject_secrets, subject_issuer)
        return subject_issuer, proof

    @staticmethod
    def create_identifier(public_bytes: bytes) -> str:
        """
        Create a new decentralised identifier.
        :param public_bytes: public key as bytes
        :return: decentralised identifier
        """
        return make_identifier(public_bytes)

    @staticmethod
    def validate_document_proof(doc: RegisterDocument):
        """
        Validate a register document proof.
        :param doc: register document

        :raises:
            IdentityInvalidDocumentError: if register document proof is invalid
        """
        DocumentValidation.validate_new_document_proof(doc)

    @staticmethod
    def create_seed(length: Optional[int] = 256) -> bytes:
        """
        Create a new seed (secrets).
        :param length: seed length
        :return: seed

        :raises:
            IdentityValidationError: if invalid seed length
        """
        if length not in (128, 256):
            raise IdentityValidationError('length must be 128 or 256')
        return secrets.token_bytes(nbytes=int(length / 8))


class AdvancedIdentityRegisterApi:
    def __init__(self, resolver_client: ResolverClient):
        self.resolver_client = resolver_client
        self.public_key_api = RegisterPublicDocKeysApi()
        self.auth_key_api = RegisterAuthPublicDocKeysApi()
        self.ctrl_deleg_api = RegisterCtrlDelegPublicDocKeysApi()
        self.auth_deleg_api = RegisterAuthDelegPublicDocKeysApi()

    def _update_doc(self, doc_owner_key_pair: KeyPair, doc_owner_issuer: Issuer, get_updated_doc: Callable, **kwargs):
        doc = self.get_register_document(doc_owner_issuer.did)
        updated_doc = get_updated_doc(doc, **kwargs)
        self.register_updated_document(updated_doc, doc_owner_key_pair, doc_owner_issuer)

    def register_updated_document(self, doc: RegisterDocument, doc_owner_key_pair: KeyPair, doc_owner_issuer: Issuer):
        """
        Register a new version of a register document against the resolver.
        :param doc: register document
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityResolverError: if resolver error
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        self.resolver_client.register_document(doc, doc_owner_key_pair.private_key, doc_owner_issuer)

    def get_register_document(self, doc_did) -> RegisterDocument:
        """
        Get a register document from the resolver.
        :param doc_did: register document decentralised identifier
        :return: associated register document

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverDocNotFoundError: if document not found
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        return self.resolver_client.get_document(doc_did)

    def get_document_if_exists(self, doc_did: str) -> Optional[RegisterDocument]:
        """
        Get a register document from the resolver if the document exists.
        :param doc_did: register document decentralised identifier
        :return: associated register document if exists else None

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        try:
            return self.get_register_document(doc_did)
        except IdentityResolverDocNotFoundError:
            return None

    def register_new_doc(self, key_pair_secrets: KeyPairSecrets, issuer: Issuer,
                         purpose: DIDType):
        """
        Create and register a new document against the resolver.
        :param key_pair_secrets: new register document owner secrets
        :param issuer: new register document owner issuer
        :param purpose: register document purpose (HOST, TWIN, USER or AGENT)

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityInvalidDocumentError: if document build error
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(key_pair_secrets)
        proof = AdvancedIdentityLocalApi.create_proof(key_pair_secrets, issuer, content=issuer.did.encode())
        doc = RegisterDocumentBuilder() \
            .add_public_key(issuer.name, key_pair.public_base58, revoked=False) \
            .build(issuer.did, purpose, proof.signature, revoked=False)
        self.register_updated_document(doc, key_pair, issuer)

    def register_new_identity_if_not_exists(self, issuer_key: IssuerKey, key_pair_secrets: KeyPairSecrets,
                                            purpose: DIDType):
        """
        Create and register a new registered identity and its associated register document against the resolver if not
        exists.
        :param issuer_key: new registered identity owner issuer key
        :param key_pair_secrets: new registered identity owner secrets
        :param purpose: registered identity purpose (HOST, TWIN, USER or AGENT)
        :return: registered identity

        :raises:
            IdentityInvalidDocumentError: if document build error
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        issuer = issuer_key.issuer
        try:
            doc = self.get_document_if_exists(issuer.did)
            if not doc:
                self.register_new_doc(key_pair_secrets, issuer_key.issuer, purpose)
            else:
                issuer = AdvancedIdentityLocalApi.get_issuer_by_public_key(doc, issuer_key.public_key_base58)
            return RegisteredIdentity(key_pair_secrets, issuer)
        except IdentityRegisterIssuerNotFoundError as err:
            raise IdentityResolverConflictError(f'Register document {issuer.did} already exist and new registered '
                                                f'identity with name {issuer.name} is not a owner') from err

    def new_registered_identity(self, purpose: DIDType, key_pair_secrets: KeyPairSecrets,
                                name: str = None, override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a new registered identity and its associated register document against the resolver.
        :param key_pair_secrets: new registered identity owner secrets
        :param name: Optional new registered identity name (default: '#<purpose>-0')
                     following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param purpose: registered identity purpose (HOST, TWIN, USER or AGENT)
        :param override_doc: override registered identity document if already exist (default False)
        :return: registered identity

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid name
            IdentityInvalidDocumentError: if document build error
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair = KeyPairSecretsHelper.get_key_pair(key_pair_secrets)
        did = AdvancedIdentityLocalApi.create_identifier(key_pair.public_bytes)
        name = name or f'{ISSUER_SEPARATOR}{purpose}-0'
        issuer = Issuer.build(did, name)
        if override_doc:
            self.register_new_doc(key_pair_secrets, issuer, purpose)
            return RegisteredIdentity(key_pair_secrets, issuer)

        issuer_key = IssuerKey.build(issuer.did, issuer.name, key_pair.public_base58)
        return self.register_new_identity_if_not_exists(issuer_key, key_pair_secrets, purpose)

    def new_registered_user_identity(self, key_pair_secrets: KeyPairSecrets,
                                     name: str = None, override_doc: bool = False):
        """
        Create and register a new user registered identity and its associated register document against the resolver.
        :param key_pair_secrets: new user registered identity owner secrets
        :param name: Optional new user registered identity name (default: '#user-0')
                     following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param override_doc: override registered identity document if already exist (default False)
        :return: user registered identity

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid name
            IdentityInvalidDocumentError: if document build error
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        return self.new_registered_identity(DIDType.USER, key_pair_secrets, name, override_doc)

    def new_registered_agent_identity(self, key_pair_secrets: KeyPairSecrets,
                                      name: str = None, override_doc: bool = False):
        """
        Create and register a new agent registered identity and its associated register document against the resolver.
        :param key_pair_secrets: new agent registered identity owner secrets
        :param name: Optional new agent registered identity name (default: '#agent-0')
                     following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param override_doc: override registered identity document if already exist (default False)
        :return: agent registered identity

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid name
            IdentityInvalidDocumentError: if document build error
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        return self.new_registered_identity(DIDType.AGENT, key_pair_secrets, name, override_doc)

    def new_registered_twin_identity(self, key_pair_secrets: KeyPairSecrets,
                                     name: str = None,
                                     override_doc: bool = False):
        """
        Create and register a new twin registered identity and its associated register document against the resolver.
        :param key_pair_secrets: new twin registered identity owner secrets
        :param name: Optional new twin registered identity name (default: '#twin-0')
                     following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param override_doc: override registered identity document if already exist (default False)
        :return: twin registered identity

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid name
            IdentityInvalidDocumentError: if document build error
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        return self.new_registered_identity(DIDType.TWIN, key_pair_secrets, name, override_doc)

    def validate_register_document(self, doc: RegisterDocument):
        """
        Validate a register document against the resolver.
        :param doc: register document

        :raises:
            IdentityInvalidDocumentDelegationError: if one of the register document delegation proof is invalid
        """
        DocumentValidation.validate_document_against_resolver(self.resolver_client, doc)

    def set_document_controller(self, doc_owner_key_pair: KeyPair,
                                doc_owner_issuer: Issuer,
                                controller: Issuer):
        """
        Set register document controller issuer.
        :param controller: register document controller issuer
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer,
                         get_updated_doc=RegisterDocumentBuilder().build_from_existing, controller=controller)

    def set_document_creator(self, doc_owner_key_pair: KeyPair,
                             doc_owner_issuer: Issuer,
                             creator: str):
        """
        Set register document creator.
        :param creator: register document creator decentralised identifier
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid creator decentralised identifier
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer,
                         get_updated_doc=RegisterDocumentBuilder().build_from_existing, creator=creator)

    def set_document_revoked(self, doc_owner_key_pair: KeyPair,
                             doc_owner_issuer: Issuer,
                             revoked: bool):
        """
        Set register document revoke field.
        :param revoked: is register document revoked
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer,
                         get_updated_doc=RegisterDocumentBuilder().build_from_existing, revoked=revoked)

    def add_public_key_to_document(self, name: str,
                                   new_public_key_base58: str,
                                   doc_owner_key_pair: KeyPair,
                                   doc_owner_issuer: Issuer) -> Issuer:
        """
        Add a new register public key to a register document.
        :param name: new public key name
        :param new_public_key_base58: public key base 58
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer
        :return: new register document owner issuer

        :raises:
            IdentityValidationError: if invalid new public key name
            IdentityRegisterDocumentKeyConflictError: if public key name is not unique within the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        new_key = RegisterPublicKey.build(name, new_public_key_base58)
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.public_key_api.add_doc_key,
                         key=new_key)
        return Issuer.build(doc_owner_issuer.did, name)

    def remove_public_key_from_document(self, removed_doc_owner_issuer: Issuer,
                                        existing_doc_owner_key_pair: KeyPair,
                                        existing_doc_owner_issuer: Issuer) -> Issuer:
        """
        Remove a register public key from a register document.
        :param removed_doc_owner_issuer: register document owner issuer to remove
        :param existing_doc_owner_key_pair: other existing register document owner key pair
        :param existing_doc_owner_issuer: other existing register document owner issuer
        :return: removed register document owner issuer

        :raises:
            IdentityValidationError: if invalid public key name
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        doc = self.get_register_document(removed_doc_owner_issuer.did)
        updated_doc = self.public_key_api.remove_doc_key(doc, key_name=removed_doc_owner_issuer.name)
        self.register_updated_document(updated_doc, existing_doc_owner_key_pair, existing_doc_owner_issuer)
        return removed_doc_owner_issuer

    def revoke_public_key_from_document(self, name: str, revoked: bool,
                                        doc_owner_key_pair: KeyPair,
                                        doc_owner_issuer: Issuer) -> Issuer:
        """
        Set register public key revoke field.
        :param name: public key name
        :param revoked: is revoked
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer
        :return: register document owner issuer

        :raises:
            IdentityValidationError: if invalid public key name
            IdentityRegisterDocumentKeyNotFoundError: if register public key not found
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.public_key_api.revoke_doc_key,
                         key_name=name, revoked=revoked)
        return Issuer.build(doc_owner_issuer.did, name)

    def add_authentication_key_to_document(self, name: str,
                                           new_public_key_base58: str,
                                           doc_owner_key_pair: KeyPair,
                                           doc_owner_issuer: Issuer) -> Issuer:
        """
        Add a new register authentication public key to a register document.
        :param name: new authentication public key name
        :param new_public_key_base58: public key base 58
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer
        :return: new key issuer

        :raises:
            IdentityValidationError: if invalid new authentication public key name
            IdentityRegisterDocumentKeyConflictError: if authentication public key name is not unique within
                                                        the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        new_key = RegisterPublicKey.build(name, new_public_key_base58)
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_key_api.add_doc_key,
                         key=new_key)
        return Issuer.build(doc_owner_issuer.did, name)

    def remove_authentication_key_from_document(self, name: str,
                                                doc_owner_key_pair: KeyPair,
                                                doc_owner_issuer: Issuer) -> Issuer:

        """
        Remove a register authentication public key from a register document.
        :param name: authentication public key name
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer
        :return: new key issuer

        :raises:
            IdentityValidationError: if invalid authentication public key name
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_key_api.remove_doc_key,
                         key_name=name)
        return Issuer.build(doc_owner_issuer.did, name)

    def revoke_authentication_key_from_document(self, name: str, revoked: bool,
                                                doc_owner_key_pair: KeyPair,
                                                doc_owner_issuer: Issuer) -> Issuer:
        """
        Set register authentication public key revoke field.
        :param name: authentication public key name
        :param revoked: is revoked
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer
        :return: new key issuer

        :raises:
            IdentityValidationError: if invalid authentication public key name
            IdentityRegisterDocumentKeyNotFoundError: if register authentication public key not found
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_key_api.revoke_doc_key,
                         key_name=name, revoked=revoked)
        return Issuer.build(doc_owner_issuer.did, name)

    def add_authentication_delegation_proof_to_document(self, proof: APIProof,
                                                        subject_issuer: Issuer,
                                                        delegation_name: str,
                                                        doc_owner_issuer: Issuer,
                                                        doc_owner_key_pair: KeyPair):
        """
        Add register authentication delegation proof to a register document.
        :param proof: proof
        :param subject_issuer: subject/delegated issuer
        :param delegation_name: register authentication delegation proof name
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register authentication delegation proof name
            IdentityRegisterDocumentKeyConflictError: if authentication delegation proof name is not unique within
                                                        the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        key = RegisterDelegationProof.build(delegation_name, subject_issuer, proof.signature,
                                            p_type=get_proof_type(proof))
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_deleg_api.add_doc_key, key=key)

    def remove_authentication_delegation_proof_from_document(self, delegation_name: str,
                                                             doc_owner_issuer: Issuer,
                                                             doc_owner_key_pair: KeyPair):
        """
        Remove register authentication delegation proof from a register document.
        :param delegation_name: register authentication delegation proof name
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register authentication delegation proof name
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_deleg_api.remove_doc_key,
                         key_name=delegation_name)

    def revoke_authentication_delegation_proof_from_document(self, delegation_name: str,
                                                             revoked: bool,
                                                             doc_owner_issuer: Issuer,
                                                             doc_owner_key_pair: KeyPair):
        """
        Set register authentication delegation proof revoke field.
        :param delegation_name: register authentication delegation proof name
        :param revoked: is revoked
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register authentication delegation proof name
            IdentityRegisterDocumentKeyNotFoundError: if register authentication delegation proof not found
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.auth_deleg_api.revoke_doc_key,
                         key_name=delegation_name, revoked=revoked)

    def add_control_delegation_proof_to_document(self, proof: APIProof,
                                                 subject_issuer: Issuer,
                                                 delegation_name: str,
                                                 doc_owner_issuer: Issuer,
                                                 doc_owner_key_pair: KeyPair):
        """
        Add register control delegation proof to a register document.
        :param proof: proof
        :param subject_issuer: subject/delegated issuer
        :param delegation_name: register control delegation proof name
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register control delegation proof name
            IdentityRegisterDocumentKeyConflictError: if control delegation proof name is not unique within
                                                       the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
       """
        key = RegisterDelegationProof.build(delegation_name, subject_issuer, proof.signature,
                                            p_type=get_proof_type(proof))
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.ctrl_deleg_api.add_doc_key, key=key)

    def remove_control_delegation_proof_from_document(self, delegation_name: str,
                                                      doc_owner_issuer: Issuer,
                                                      doc_owner_key_pair: KeyPair):
        """
        Remove register control delegation proof from a register document.
        :param delegation_name: register control delegation proof name
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register control delegation proof name
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.ctrl_deleg_api.remove_doc_key,
                         key_name=delegation_name)

    def revoke_control_delegation_proof_from_document(self, delegation_name: str,
                                                      revoked: bool,
                                                      doc_owner_issuer: Issuer,
                                                      doc_owner_key_pair: KeyPair):
        """
        Set register control delegation proof revoke field.
        :param delegation_name: register control delegation proof name
        :param revoked: is revoked
        :param doc_owner_key_pair: register document owner key pair
        :param doc_owner_issuer: register document owner issuer

        :raises:
            IdentityValidationError: if invalid register control delegation proof name
            IdentityRegisterDocumentKeyNotFoundError: if register control delegation proof not found
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        self._update_doc(doc_owner_key_pair, doc_owner_issuer, get_updated_doc=self.ctrl_deleg_api.revoke_doc_key,
                         key_name=delegation_name, revoked=revoked)

    def _delegate(self, delegating_secrets: KeyPairSecrets, delegating_did: str,
                  subject_secrets: KeyPairSecrets, subject_did: str, delegation_name: str,
                  proof_type: DelegationProofType, delegation_proof_add_method: Callable):
        delegating_doc = self.get_register_document(delegating_did)
        subject_doc = self.get_register_document(subject_did)
        delegating_key_pair = KeyPairSecretsHelper.get_key_pair(delegating_secrets)
        delegating_issuer = AdvancedIdentityLocalApi.get_issuer_by_public_key(delegating_doc,
                                                                              delegating_key_pair.public_base58)
        if proof_type == DelegationProofType.GENERIC:
            subject_issuer, proof = AdvancedIdentityLocalApi.create_generic_delegation_proof(subject_doc,
                                                                                             subject_secrets)
        else:
            subject_issuer, proof = AdvancedIdentityLocalApi.create_delegation_proof(delegating_issuer,
                                                                                     subject_doc,
                                                                                     subject_secrets)

        existing_delegation = RegisterDocumentHelper.get_register_delegation_proof_by_controller(subject_issuer,
                                                                                                 delegating_doc,
                                                                                                 True)
        if existing_delegation and (
                not delegation_name or delegation_name == existing_delegation.name
        ):
            # Found an existing delegation with matching controller and name. Nothing to do.
            return

        if not delegation_name:
            delegation_name = RegisterDocumentHelper.new_random_name_for_document(subject_doc)

        delegation_proof_add_method(proof, subject_issuer, delegation_name,
                                    delegating_issuer, delegating_key_pair)

    def delegate_authentication(self, delegating_secrets: KeyPairSecrets, delegating_did: str,
                                subject_secrets: KeyPairSecrets, subject_did: str, delegation_name: str = None,
                                proof_type: DelegationProofType = DelegationProofType.DID):
        """
        Delegate authentication between delegating registered identity and delegated registered identity.
        :param delegating_secrets: delegating identity secrets
        :param delegating_did: delegating identity decentralised identifier
        :param subject_secrets: subject/delegated identity secrets
        :param subject_did: subject/delegated identity decentralised identifier
        :param delegation_name: Optional delegation name. If None a random name will be chosen
        :param proof_type: Delegation proof type. By default: set to DID, that means the proof can be used to setup a
        delegation from single delegating issuer. GENERIC means the proof can be used to setup a delegation from
         several delegating issuers.

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid register authentication delegation proof name
            IdentityRegisterIssuerNotFoundError: if the delegating (or delegated) issuer matching the delegating
                                                (or delegated) secrets is not found
            IdentityRegisterDocumentKeyConflictError: if authentication delegation proof name is not unique within
                                                      the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
            IdentityDependencyError: if incompatible library dependency
        """
        self._delegate(delegating_secrets,
                       delegating_did,
                       subject_secrets,
                       subject_did,
                       delegation_name,
                       proof_type,
                       delegation_proof_add_method=self.add_authentication_delegation_proof_to_document)

    def delegate_control(self, delegating_secrets: KeyPairSecrets, delegating_did: str, subject_secrets: KeyPairSecrets,
                         subject_did: str, delegation_name: str = None,
                         proof_type: DelegationProofType = DelegationProofType.DID):
        """
        Delegate control between delegating registered identity and delegated registered identity.
        :param delegating_secrets: delegating identity secrets
        :param delegating_did: delegating identity decentralised identifier
        :param subject_secrets: subject/delegated identity secrets
        :param subject_did: subject/delegated identity decentralised identifier
        :param delegation_name: Optional delegation name. If None a random name will be chosen
        :param proof_type: Delegation proof type. By default: set to DID, that means the proof can be used to setup a
        delegation from single delegating issuer. GENERIC means the proof can be used to setup a delegation from
         several delegating issuers.

        :raises:
            IdentityValidationError: if invalid secrets
            IdentityValidationError: if invalid register control delegation proof name
            IdentityRegisterIssuerNotFoundError: if the delegating (or delegated) issuer matching the delegating
                                                (or delegated) secrets is not found
            IdentityRegisterDocumentKeyConflictError: if control delegation proof name is not unique within
                                                      the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
            IdentityDependencyError: if incompatible library dependency
        """
        self._delegate(delegating_secrets,
                       delegating_did,
                       subject_secrets,
                       subject_did,
                       delegation_name,
                       proof_type,
                       delegation_proof_add_method=self.add_control_delegation_proof_to_document)

    def get_proof_from_challenge_token(self, challenge_token: str) -> Proof:
        """
        Get the proof from a challenge token.
        :param challenge_token: challenge jwt token
        :return: valid proof

        :raises:
            IdentityValidationError: if invalid challenge token
        """
        return Proof.from_challenge_token(self.resolver_client, challenge_token)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from _blake2 import blake2b  # type: ignore

import base58

from iotics.lib.identity.const import IDENTIFIER_PREFIX, ISSUER_SEPARATOR


def is_same_identifier(issuer_a: str, issuer_b: str) -> bool:
    """
    check if 2 issuers string have the same identifier
    :param issuer_a: issuer string a
    :param issuer_b: issuer string b
    :return: True if identifier are the same else False
    """
    return issuer_a.split(ISSUER_SEPARATOR)[0] == issuer_b.split(ISSUER_SEPARATOR)[0]


IDENTIFIER_METHOD = 0x05
IDENTIFIER_VERSION = 0x55
IDENTIFIER_PAD = 0x59


def make_identifier(public_bytes: bytes) -> str:
    """
    Generate a new decentralised identifier from public key as bytes
    :param public_bytes: public key as bytes
    :return: decentralised identifier
    """
    bl2 = blake2b(digest_size=20)
    bl2.update(public_bytes)
    pk_digest = bl2.digest()

    cl2 = blake2b(digest_size=20)
    cl2.update(pk_digest)
    checksum = bytearray.fromhex(cl2.hexdigest())[:4]

    return IDENTIFIER_PREFIX + base58.b58encode(bytes([IDENTIFIER_METHOD, IDENTIFIER_VERSION, IDENTIFIER_PAD])
                                                + pk_digest + checksum).decode('ascii')



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from typing import Optional, Tuple, Union

from iotics.lib.identity.api.advanced_api import AdvancedIdentityLocalApi, AdvancedIdentityRegisterApi
from iotics.lib.identity.crypto.key_pair_secrets import build_agent_secrets, build_twin_secrets, build_user_secrets, \
    KeyPairSecretsHelper, SeedMethod
from iotics.lib.identity.register.key_pair import RegisteredIdentity
from iotics.lib.identity.register.rest_resolver import get_rest_resolver_client


class HighLevelIdentityApi:
    def __init__(self, advanced_api: AdvancedIdentityRegisterApi):
        self.advanced_api = advanced_api

    def create_user_and_agent_with_auth_delegation(self, user_seed: bytes, user_key_name: str,
                                                   agent_seed: bytes, agent_key_name: str,
                                                   delegation_name: str = None,
                                                   user_name: str = None, agent_name: str = None,
                                                   user_password: str = '', agent_password: str = '',
                                                   override_docs: bool = False, ) -> Tuple[RegisteredIdentity,
                                                                                           RegisteredIdentity]:
        """
        Create and register a user and an agent identities with user delegating authentication to the agent.
        The agent can authenticate on behalf of the user.
        :param user_seed: user seed (secrets)
        :param user_key_name: user key name (secrets)
        :param user_password: Optional user password (secrets)
        :param user_name: Optional user name (default #user-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param agent_seed: agent seed (secrets)
        :param agent_key_name: agent key name (secrets)
        :param agent_password: Optional agent password (secrets)
        :param agent_name: Optional agent name (default #agent-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param delegation_name: Optional delegation name. If None a random name will be chosen
        :param override_docs: override registered identity documents if already exist (default False)
        :return: user registered identity, agent registered identity

        :raises:
            IdentityValidationError: if invalid secrets or names
            IdentityValidationError: if invalid names
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        user_key_pair_secrets = build_user_secrets(user_seed, user_key_name, SeedMethod.SEED_METHOD_BIP39,
                                                   user_password)
        user_registered_identity = self.advanced_api.new_registered_user_identity(user_key_pair_secrets, user_name,
                                                                                  override_docs)
        agent_key_pair_secrets = build_agent_secrets(agent_seed, agent_key_name, SeedMethod.SEED_METHOD_BIP39,
                                                     agent_password)
        agent_registered_identity = self.advanced_api.new_registered_agent_identity(agent_key_pair_secrets,
                                                                                    agent_name,
                                                                                    override_docs)
        self.advanced_api.delegate_authentication(user_registered_identity.key_pair_secrets,
                                                  user_registered_identity.issuer.did,
                                                  agent_registered_identity.key_pair_secrets,
                                                  agent_registered_identity.issuer.did,
                                                  delegation_name)
        return user_registered_identity, agent_registered_identity

    @staticmethod
    def create_agent_auth_token(agent_registered_identity: RegisteredIdentity,
                                user_did: str, duration: int, audience: str = 'default') -> str:
        """
        Create an agent authentication token.
        :param agent_registered_identity: agent registered identity
        :param user_did: user register document decentralised identifier
        :param duration: token duration in seconds
        :param audience: Optional token audience
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if invalid agent identity
            IdentityValidationError: if invalid token data
            IdentityDependencyError: if incompatible library dependency
        """
        agent_key_pair = KeyPairSecretsHelper.get_key_pair(agent_registered_identity.key_pair_secrets)
        return AdvancedIdentityLocalApi.create_agent_auth_token(agent_key_pair,
                                                                agent_registered_identity.issuer,
                                                                user_did,
                                                                duration,
                                                                audience)

    def create_twin(self, twin_seed: bytes, twin_key_name: str, twin_name: str = None,
                    password: str = '', override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a twin identity.
        :param twin_seed: twin seed (secrets)
        :param twin_key_name: twin key name (secrets)
        :param password: Optional twin password (secrets)
        :param twin_name: Optional twin name (default #twin-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param override_doc: override registered identity document if already exist (default False)
        :return: twin registered identity

        :raises:
            IdentityValidationError: if invalid secrets or name
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        key_pair_secrets = build_twin_secrets(twin_seed, twin_key_name, SeedMethod.SEED_METHOD_BIP39, password)
        return self.advanced_api.new_registered_twin_identity(key_pair_secrets, twin_name, override_doc)

    def create_twin_with_control_delegation(self, twin_seed: bytes, twin_key_name: str,
                                            agent_registered_identity: RegisteredIdentity,
                                            delegation_name: str = None,
                                            twin_name: str = None,
                                            password: str = '',
                                            override_doc: bool = False) -> RegisteredIdentity:
        """
        Create and register a twin identity with twin delegating control to the agent.
        The agent can control the twin.
        :param twin_seed: twin seed (secrets)
        :param twin_key_name: twin key name (secrets)
        :param agent_registered_identity: Agent identity
        :param delegation_name: Optional delegation name. If None a random name will be chosen
        :param twin_name: Optional twin name (default #twin-0) following this pattern: '#[a-zA-Z\\-\\_0-9]{1, 24}'
        :param override_doc: override registered identity document if already exist (default False)
        :return: twin registered identity

        :raises:
            IdentityValidationError: if invalid secret or name
            IdentityResolverConflictError: register document already exists with different owners
            IdentityResolverError: if can not interact with the resolver
            IdentityDependencyError: if incompatible library dependency
        """
        twin_registered_identity = self.create_twin(twin_seed, twin_key_name, twin_name, password, override_doc)
        self.advanced_api.delegate_control(twin_registered_identity.key_pair_secrets,
                                           twin_registered_identity.issuer.did,
                                           agent_registered_identity.key_pair_secrets,
                                           agent_registered_identity.issuer.did,
                                           delegation_name)

        return twin_registered_identity

    def get_ownership_of_twin_from_registered_identity(self, twin_registered_identity: RegisteredIdentity,
                                                       new_owner_registered_identity: RegisteredIdentity,
                                                       new_owner_key_name: str):
        """
        Get Ownership of a twin using a registered identity you owned.
        :param twin_registered_identity: twin registered identity
        :param new_owner_registered_identity: new owner registered identity
        :param new_owner_key_name: new owner key name

        :raises:
            IdentityValidationError: if invalid registered identities or name
            IdentityRegisterDocumentKeyConflictError: if new owner key name is not unique within the register document
            IdentityInvalidDocumentError: if invalid register document
            IdentityResolverError: if resolver error
        """
        new_owner_key_pair = KeyPairSecretsHelper.get_key_pair(new_owner_registered_identity.key_pair_secrets)
        current_owner_key_pair = KeyPairSecretsHelper.get_key_pair(twin_registered_identity.key_pair_secrets)
        self.advanced_api.add_public_key_to_document(new_owner_key_name, new_owner_key_pair.public_base58,
                                                     current_owner_key_pair, twin_registered_identity.issuer)

    @staticmethod
    def create_seed(length: Optional[int] = 256) -> bytes:
        """
        Create a new seed (secrets).
        :param length: seed length
        :return: seed

        :raises:
            IdentityValidationError: if invalid seed length
        """
        return AdvancedIdentityLocalApi.create_seed(length)


def get_rest_high_level_identity_api(resolver_url, timeout: Optional[Union[int, float]] = None) -> HighLevelIdentityApi:
    """
    Get a REST high level identity api.
    :param resolver_url: resolver url
    :param timeout: optional timeout seconds. Default=60s. If set to 0, requests will have no timeout.
    :return: high level identity api
    """
    resolver_client = get_rest_resolver_client(resolver_url, timeout)
    return HighLevelIdentityApi(AdvancedIdentityRegisterApi(resolver_client))



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.proof import Proof
from iotics.lib.identity.error import IdentityInvalidDocumentError, IdentityInvalidProofError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.document_helper import RegisterDocumentHelper
from iotics.lib.identity.register.resolver import ResolverClient
from iotics.lib.identity.validation.proof import DelegationValidation, ProofValidation


class DocumentValidation:

    @staticmethod
    def validate_new_document_proof(doc: RegisterDocument):
        """
        Validate register document proof.
        :param doc: register document.

        :raises:
            IdentityInvalidDocumentError: if register document initial owner public key has been removed
            IdentityInvalidDocumentError: if register document proof is invalid
        """
        try:
            key = RegisterDocumentHelper.get_owner_register_public_key(doc)
            if not key:
                raise IdentityInvalidDocumentError(f'Invalid document \'{doc.did}\', no owner public key')
            ProofValidation.validate_proof(
                Proof(Issuer.build(doc.did, key.name), doc.did.encode('ascii'), doc.proof), key.base58)
        except IdentityInvalidProofError as err:
            raise IdentityInvalidDocumentError(f'Invalid document \'{doc.did}\' proof: {err}') from err

    @staticmethod
    def validate_document_against_resolver(resolver_client: ResolverClient, doc: RegisterDocument):
        """
        Validate a register document against the resolver.
        :param resolver_client: resolver client interface
        :param doc: register document

        :raises:
            IdentityInvalidDocumentDelegationError: if one of the register document delegation proof is invalid
        """
        for key in doc.control_delegation_proof.values():
            DelegationValidation.validate_delegation(resolver_client, doc.did, key)

        for key in doc.auth_delegation_proof.values():
            DelegationValidation.validate_delegation(resolver_client, doc.did, key)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.
from iotics.lib.identity.crypto.identity import is_same_identifier
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.jwt import JwtTokenHelper
from iotics.lib.identity.error import IdentityAuthenticationFailed, IdentityInvalidDocumentDelegationError, \
    IdentityInvalidRegisterIssuerError, IdentityNotAllowed, IdentityResolverError, IdentityValidationError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.document_helper import RegisterDocumentHelper
from iotics.lib.identity.register.resolver import ResolverClient
from iotics.lib.identity.validation.proof import DelegationValidation


class IdentityAuthValidation:
    @staticmethod
    def is_allowed_for(issuer: Issuer, issuer_doc: RegisterDocument, subject_doc: RegisterDocument,
                       include_auth: bool) -> bool:
        """
        Check if the issuer is allowed for control (authentication if include_auth = True) on the subject register
        document.
        Issuer is allowed if both the issuer and subject register document are not revoked
        AND (
             ( the issuer is the owner of the subject register document
             OR
              if a include_auth=True the issuer is in the authentication public keys of the subject register document
             )
             OR
             the issuer is delegated for control (authentication if include_auth = True) with a valid delegation proof
             on the subject registered document
        )

        :param issuer: issuer
        :param issuer_doc: issuer register document
        :param subject_doc: subject register document
        :param include_auth: include authentication keys and delegation proof is set to True
        :return: True is allowed else False
        """
        if issuer_doc.revoked or subject_doc.revoked:
            return False

        if is_same_identifier(issuer.did, subject_doc.did):  # it is the same document
            issuer_key = RegisterDocumentHelper.get_issuer_register_key(issuer.name, subject_doc, include_auth)
            if issuer_key and not issuer_key.revoked:
                return True

        delegation_proof = RegisterDocumentHelper.get_register_delegation_proof_by_controller(issuer, subject_doc,
                                                                                              include_auth)
        if delegation_proof:
            try:
                DelegationValidation.validate_delegation_from_doc(subject_doc.did, issuer_doc, delegation_proof)
            except IdentityInvalidDocumentDelegationError:
                return False
            return not delegation_proof.revoked
        return False

    @staticmethod
    def _check_allowed_on_doc_or_controller(resolver_client: ResolverClient, issuer: Issuer, subject_id: str,
                                            include_auth: bool):
        """
        Validate if issuer is allowed for control  (authentication if include_auth = True) on the register document
        associated to the subject decentralised identifier.
        Issuer is allowed if both the issuer and subject register document can be fetched
        AND (
            if the issuer is allowed on the subject register document
            OR
            if the issuer is allowed on the subject controller register document
        )

        :param resolver_client: resolver client interface
        :param issuer: issuer under validation
        :param subject_id: subject register document decentralised identifier
        :param include_auth: include authentication keys and delegation proof is set to True

        :raises:
             IdentityNotAllowed: if issuer not allowed
        """
        try:
            issuer_doc = resolver_client.get_document(issuer.did)
            subject_doc = resolver_client.get_document(subject_id)
            if IdentityAuthValidation.is_allowed_for(issuer, issuer_doc, subject_doc, include_auth):
                return
            # Check if allowed for controller
            if subject_doc.controller:
                controller_doc = resolver_client.get_document(subject_doc.controller.did)
                if IdentityAuthValidation.is_allowed_for(issuer, issuer_doc, controller_doc, include_auth):
                    return
        except IdentityResolverError as err:
            raise IdentityNotAllowed(f'Cannot validate issuer {issuer} is allowed for {subject_id}: {err}') from err
        raise IdentityNotAllowed(f'Issuer {issuer} not allowed for {subject_id}')

    @staticmethod
    def validate_allowed_for_control(resolver_client: ResolverClient, issuer: Issuer, subject_id: str):
        """
        Validate if issuer is allowed for control on the register document associated to the subject decentralised
        identifier.
        :param resolver_client: resolver client interface
        :param issuer: issuer under validation
        :param subject_id: subject register document decentralised identifier

        :raises:
            IdentityNotAllowed: if the issuer is not allowed for control
        """
        IdentityAuthValidation._check_allowed_on_doc_or_controller(resolver_client, issuer, subject_id,
                                                                   include_auth=False)

    @staticmethod
    def validate_allowed_for_auth(resolver_client: ResolverClient, issuer: Issuer, subject_id: str):
        """
        Validate if issuer is allowed for authentication on the register document associated to the subject
        decentralised identifier.
        :param resolver_client: resolver client interface
        :param issuer: issuer under validation
        :param subject_id: subject register document decentralised identifier

        :raises:
            IdentityNotAllowed: if the issuer is not allowed for authentication
        """
        IdentityAuthValidation._check_allowed_on_doc_or_controller(resolver_client, issuer, subject_id,
                                                                   include_auth=True)

    @staticmethod
    def verify_authentication(resolver_client: ResolverClient, token: str) -> dict:
        """
        Verify if the authentication token is allowed for authentication.
        :param resolver_client: resolver client interface
        :param token: jwt authentication token
        :return: decoded verified authentication token

        :raises:
            IdentityAuthenticationFailed: if not allowed for authentication
        """
        try:
            unverified_token = JwtTokenHelper.decode_token(token)
            for field in ('iss', 'sub', 'aud', 'iat', 'exp'):
                if field not in unverified_token:
                    raise IdentityValidationError(f'Invalid token, missing {field} field')
            issuer = Issuer.from_string(unverified_token['iss'])
            doc = resolver_client.get_document(issuer.did)
            get_controller_doc = resolver_client.get_document
            issuer_key = RegisterDocumentHelper.get_valid_issuer_key_for_auth(doc, issuer.name, get_controller_doc)
            if not issuer_key:
                raise IdentityInvalidRegisterIssuerError(f'Invalid issuer {issuer}')
            verified_token = JwtTokenHelper.decode_and_verify_token(token, issuer_key.public_key_base58,
                                                                    unverified_token['aud'])

            IdentityAuthValidation.validate_allowed_for_auth(resolver_client, issuer_key.issuer, verified_token['sub'])

            return {'iss': verified_token['iss'],
                    'sub': verified_token['sub'],
                    'aud': verified_token['aud'],
                    'iat': verified_token['iat'],
                    'exp': verified_token['exp']}
        except (IdentityValidationError, IdentityResolverError,
                IdentityInvalidRegisterIssuerError, IdentityNotAllowed) as err:
            raise IdentityAuthenticationFailed('Not authenticated') from err



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from time import time as unixtime
from typing import Dict, Optional, Set

from iotics.lib.identity.const import DOCUMENT_VERSION, SUPPORTED_VERSIONS
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.key_pair_secrets import DIDType
from iotics.lib.identity.error import IdentityInvalidDocumentError, IdentityRegisterDocumentKeyConflictError
from iotics.lib.identity.register.document import Metadata, RegisterDocument
from iotics.lib.identity.register.keys import RegisterAuthenticationPublicKey, RegisterDelegationProof, \
    RegisterPublicKey
from iotics.lib.identity.validation.identity import IdentityValidation


def get_unix_time_ms() -> int:
    return int(unixtime() * 1000)


class RegisterDocumentBuilder:
    def __init__(self):
        self.public_keys: Dict[str, RegisterPublicKey] = {}
        self.auth_keys: Dict[str, RegisterAuthenticationPublicKey] = {}
        self.control_delegations: Dict[str, RegisterDelegationProof] = {}
        self.auth_delegations: Dict[str, RegisterDelegationProof] = {}
        self.name_set: Set[str] = set()
        self.revoked: bool = False

    def _check_and_update_map(self, name: str, val: object, key_map: dict, message_prefix: str):
        if name in self.name_set and not val.is_equal(key_map.get(name)):  # type: ignore
            raise IdentityRegisterDocumentKeyConflictError(f'{message_prefix} \'{name}\' already in use')

        self.name_set.add(name)
        key_map[name] = val

    def _remove_key(self, name: str):
        self.name_set.discard(name)
        self.public_keys.pop(name, None)
        self.auth_keys.pop(name, None)
        self.control_delegations.pop(name, None)
        self.auth_delegations.pop(name, None)

    def add_public_key(self, name: str, public_base58: str,
                       revoked: Optional[bool] = False) -> 'RegisterDocumentBuilder':
        """
        Add public key to document under build.
        :param name: key name
        :param public_base58: public key base58
        :param revoked: is revoked (default=False)
        :return: self

        :raises:
            IdentityValidationError: if invalid public key
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        key = RegisterPublicKey.build(name, public_base58, revoked)
        self._check_and_update_map(name, key, self.public_keys, 'Public key name')
        return self

    def add_public_key_obj(self, key: RegisterPublicKey) -> 'RegisterDocumentBuilder':
        """
        Add public key to document under build.
        :param key: register public key
        :return: self

        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        self._check_and_update_map(key.name, key, self.public_keys, 'Public key name')
        return self

    def add_authentication_key(self, name: str, public_base58: str,
                               revoked: Optional[bool] = False) -> 'RegisterDocumentBuilder':
        """
        Add authentication public key to document under build.
        :param name: key name
        :param public_base58: authentication public key base58
        :param revoked: is revoked (default=False)
        :return: self

        :raises:
            IdentityValidationError: if invalid authentication public key
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        key = RegisterAuthenticationPublicKey.build(name, public_base58, revoked)
        self._check_and_update_map(name, key, self.auth_keys, 'Authentication key name')
        return self

    def add_authentication_key_obj(self, key: RegisterAuthenticationPublicKey) -> 'RegisterDocumentBuilder':
        """
        Add authentication public key to document under build.
        :param key: register authentication public key
        :return: self

        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        self._check_and_update_map(key.name, key, self.auth_keys, 'Authentication key name')
        return self

    def add_control_delegation(self, name: str, controller: Issuer, proof: str,
                               revoked: Optional[bool] = False) -> 'RegisterDocumentBuilder':
        """
        Add control delegation public key to document under build.
        :param name: key name
        :param controller: control delegation controller
        :param proof: control delegation proof
        :param revoked: is revoked (default=False)
        :return: self

        :raises:
            IdentityValidationError: if invalid control delegation public key
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        key = RegisterDelegationProof.build(name, controller, proof, revoked)
        self._check_and_update_map(name, key, self.control_delegations, 'Control delegation name')
        return self

    def add_control_delegation_obj(self, key: RegisterDelegationProof) -> 'RegisterDocumentBuilder':
        """
        Add control delegation public key to document under build.
        :param key: register control delegation public key
        :return: self

        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        self._check_and_update_map(key.name, key, self.control_delegations, 'Control delegation name')
        return self

    def add_authentication_delegation(self, name: str, controller: Issuer, proof: str,
                                      revoked: Optional[bool] = False) -> 'RegisterDocumentBuilder':
        """
        Add authentication delegation public key to document under build.
        :param name: key name
        :param controller: authentication delegation controller
        :param proof: authentication delegation proof
        :param revoked: is revoked (default=False)
        :return: self

        :raises:
            IdentityValidationError: if invalid authentication delegation public key
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        key = RegisterDelegationProof.build(name, controller, proof, revoked)
        self._check_and_update_map(name, key, self.auth_delegations,
                                   'Authentication delegation name')
        return self

    def add_authentication_delegation_obj(self, key: RegisterDelegationProof) -> 'RegisterDocumentBuilder':
        """
        Add authentication delegation public key to document under build.
        :param key: register authentication delegation public key
        :return: self

        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        self._check_and_update_map(key.name, key, self.auth_delegations,
                                   'Authentication delegation name')
        return self

    def remove_key(self, key_name: str) -> 'RegisterDocumentBuilder':
        """
        Remove a key from the doc under build. Do nothing if the key do not
        belongs to the doc.
        :param key_name: the key name
        :return: self
        """
        self._remove_key(key_name)
        return self

    def set_keys_from_existing(self, doc: RegisterDocument) -> 'RegisterDocumentBuilder':
        """
        Set keys from a valid immutable register document from an existing one.
        :param doc: existing register document
        :return: self

        :raises:
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
       """
        for pub_key in doc.public_keys.values():
            self.add_public_key_obj(pub_key)
        for auth_key in doc.auth_keys.values():
            self.add_authentication_key_obj(auth_key)
        for auth_proof in doc.auth_delegation_proof.values():
            self.add_authentication_delegation_obj(auth_proof)
        for ctrl_proof in doc.control_delegation_proof.values():
            self.add_control_delegation_obj(ctrl_proof)
        return self

    def build(self, did: str, purpose: DIDType, proof: str, revoked: bool,
              metadata: Metadata = Metadata(), creator: Optional[str] = None,
              spec_version: Optional[str] = DOCUMENT_VERSION,
              update_time: Optional[int] = None,
              controller: Issuer = None) -> RegisterDocument:
        """
        Build a valid immutable register document.
        :param did: register document decentralised identifier
        :param purpose: register document purpose (HOST, TWIN, USER or AGENT)
        :param proof: register document proof
        :param revoked: is register document revoked
        :param metadata: register document optional metadata.
        :param creator: register document optional creator identifier.
        :param spec_version: register document version (default=DOCUMENT_VERSION)
        :param update_time: register document update time (unix time ms). Automatically set to know if not provided.
        :param controller: register document optional controller
        :return: valid register document

        :raises:
            IdentityInvalidDocumentError: if version ot supported
            IdentityValidationError: if invalid controller
            IdentityInvalidDocumentError: if invalid document (no controller and no public key)
        """
        IdentityValidation.validate_identifier(did)
        if spec_version not in SUPPORTED_VERSIONS:
            raise IdentityInvalidDocumentError(f'Unsupported version {spec_version} not in {SUPPORTED_VERSIONS}')
        if creator:
            IdentityValidation.validate_identifier(creator)
        if not controller and not self.public_keys:
            raise IdentityInvalidDocumentError('Invalid document, no Controller or Public Keys provided')
        return RegisterDocument(did=did, purpose=purpose, proof=proof, revoked=revoked,
                                metadata=metadata, creator=creator,
                                public_keys=self.public_keys,
                                auth_keys=self.auth_keys,
                                control_delegation_proof=self.control_delegations,
                                auth_delegation_proof=self.auth_delegations,
                                spec_version=spec_version,
                                update_time=update_time or get_unix_time_ms(),
                                controller=controller)

    def build_from_dict(self, data: dict) -> RegisterDocument:
        """
        Build a valid immutable register document from dict.
        :param data: register document as dict
        :return: valid register document

        :raises:
            IdentityInvalidDocumentError: if invalid dict data
            IdentityInvalidDocumentError: if invalid document
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        try:
            for k in data['publicKey']:
                self.add_public_key_obj(RegisterPublicKey.from_dict(k))
            for k in data.get('authentication', []):
                self.add_authentication_key_obj(RegisterAuthenticationPublicKey.from_dict(k))
            for k in data.get('delegateAuthentication', []):
                self.add_authentication_delegation_obj(RegisterDelegationProof.from_dict(k))
            for k in data.get('delegateControl', []):
                self.add_control_delegation_obj(RegisterDelegationProof.from_dict(k))
            raw_controller = data.get('controller')
            controller = Issuer.from_string(raw_controller) if raw_controller else None
            return self.build(data['id'], DIDType(data['ioticsDIDType']), data['proof'], data.get('revoked', False),
                              Metadata.from_dict(data.get('metadata', {})), data.get('creator'),
                              data['ioticsSpecVersion'], data['updateTime'], controller)
        except (TypeError, KeyError, ValueError) as err:
            raise IdentityInvalidDocumentError(f'Can not parse invalid register document: \'{err}\'') from err

    def build_from_existing(self, doc: RegisterDocument,
                            revoked: Optional[bool] = None,
                            metadata: Optional[Metadata] = None,
                            creator: Optional[str] = None,
                            spec_version: Optional[str] = None,
                            controller: Optional[Issuer] = None,
                            populate_with_doc_keys: Optional[bool] = True) -> RegisterDocument:
        """
        Build a valid immutable register document from an existing one.
        :param doc: existing register document
        :param revoked: is register document revoked. Optional, takes the existing doc value if not provided.
        :param metadata: register document optional metadata. Optional, takes the existing doc value if not provided.
        :param creator: register document optional creator identifier. Optional, takes the existing doc value
        if not provided.
        :param spec_version: register document version (default=DOCUMENT_VERSION). Optional, takes the existing doc
        value if not provided.
        :param controller: register document optional controller. Optional, takes the existing doc value
        if not provided.
        :param populate_with_doc_keys: populate builder keys with keys from the existion document (default=True)
        :return: valid register document

        :raises:
            IdentityValidationError: if invalid controller
            IdentityInvalidDocumentError: if invalid document
            IdentityRegisterDocumentKeyConflictError: if key name is not unique
        """
        if populate_with_doc_keys:
            self.set_keys_from_existing(doc)
        revoked = revoked if revoked is not None else doc.revoked
        return self.build(doc.did, doc.purpose, doc.proof,
                          revoked,
                          metadata or doc.metadata,
                          creator or doc.creator,
                          spec_version or doc.spec_version,
                          controller=controller or doc.controller)



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from datetime import datetime

import jwt
from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.const import DEFAULT_TOKEN_START_OFFSET_SECONDS, TOKEN_ALGORITHM
from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.keys import KeysHelper
from iotics.lib.identity.error import IdentityValidationError
from iotics.lib.identity.register.document import RegisterDocument


class JwtTokenHelper:

    @staticmethod
    def decode_token(token: str) -> dict:
        """
        Decode a jwt token without verifying it.
        :param token: jwt token
        :return: decoded token

        :raises:
            IdentityValidationError: if invalid token
        """
        try:
            return jwt.decode(token, options={'verify_signature': False},
                              algorithms=[TOKEN_ALGORITHM], verify=False)
        except jwt.exceptions.DecodeError as err:
            raise IdentityValidationError(f'Can not decode invalid token: \'{err}\'') from err

    @staticmethod
    def decode_and_verify_token(token: str, public_base58: str, audience: str):
        """
        Decode a jwt token and verifying it.
        :param token: jwt token
        :param public_base58:  token public base58 key
        :param audience: token audience
        :return: decoded verified token

        :raises:
            IdentityValidationError: if invalid token
            IdentityValidationError: if invalid token signature
            IdentityValidationError: if expired token
        """
        try:
            key = KeysHelper.get_public_ECDSA_from_base58(public_base58)
            return jwt.decode(token, key, audience=audience, algorithms=[TOKEN_ALGORITHM],  # type: ignore
                              verify=True, options={'verify_signature': True})
        except jwt.exceptions.InvalidSignatureError as err:
            raise IdentityValidationError(f'Invalid token signature: \'{err}\'') from err
        except jwt.exceptions.ExpiredSignatureError as err:
            raise IdentityValidationError(f'Expired token: \'{err}\'') from err
        except jwt.exceptions.DecodeError as err:
            raise IdentityValidationError(f'Can not decode invalid token: \'{err}\'') from err

    @staticmethod
    def create_doc_token(issuer: Issuer, audience: str, doc: RegisterDocument,
                         private_key: ec.EllipticCurvePrivateKey) -> str:
        """
        Create a register document jwt token.
        :param issuer: document issuer
        :param audience: token audience
        :param doc: register document
        :param private_key: issuer private key
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if can not encode the token
        """
        try:
            return jwt.encode({'iss': str(issuer), 'aud': audience, 'doc': doc.to_dict()}, private_key,  # type: ignore
                              algorithm=TOKEN_ALGORITHM)
        except (TypeError, ValueError) as err:
            raise IdentityValidationError(f'Can not create document token for {issuer}: \'{err}\'') from err

    @staticmethod
    def create_auth_token(iss: str, sub: str, aud: str, duration: int,
                          private_key: ec.EllipticCurvePrivateKey,
                          start_offset: int = DEFAULT_TOKEN_START_OFFSET_SECONDS) -> str:
        """
        Create an authentication jwt token.
        :param iss: issuer as string
        :param sub: subject document did
        :param aud: token audience
        :param duration: token duration (seconds)
        :param private_key: issuer private key
        :param start_offset: offset for token valid-from time used (default=DEFAULT_TOKEN_START_OFFSET_SECONDS)
        :return: encoded jwt token

        :raises:
            IdentityValidationError: if invalid duration (<=0)
            IdentityValidationError: if can not encode the token
        """
        now = int(datetime.now().timestamp())
        if duration < 0:
            raise IdentityValidationError(f'Can not create auth token with duration={duration}, must be >0')
        try:
            return jwt.encode({
                'iss': iss,
                'aud': aud,
                'sub': sub,
                'iat': now + start_offset,
                'exp': now + duration
            }, private_key, algorithm='ES256')  # type: ignore
        except (TypeError, ValueError) as err:
            raise IdentityValidationError(f'Can not create auth token for {iss}/{sub}: \'{err}\'') from err



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from iotics.lib.identity.api import *  # noqa: F403,F401
from iotics.lib.identity.error import *  # noqa: F403,F401
from iotics.lib.identity.register.resolver import *  # noqa: F403,F401
from iotics.lib.identity.register.rest_resolver import *  # noqa: F403,F401



# Copyright (c) IOTIC LABS LIMITED. All rights reserved. Licensed under the Apache License, Version 2.0.

from http import HTTPStatus
from typing import Optional, Union

import requests
from cryptography.hazmat.primitives.asymmetric import ec

from iotics.lib.identity.crypto.issuer import Issuer
from iotics.lib.identity.crypto.jwt import JwtTokenHelper
from iotics.lib.identity.error import IdentityInvalidRegisterIssuerError, IdentityResolverCommunicationError, \
    IdentityResolverError, IdentityResolverHttpDocNotFoundError, IdentityResolverHttpError, \
    IdentityResolverTimeoutError, IdentityValidationError
from iotics.lib.identity.register.document import RegisterDocument
from iotics.lib.identity.register.document_builder import RegisterDocumentBuilder
from iotics.lib.identity.register.document_helper import GetControllerDocFunc, RegisterDocumentHelper
from iotics.lib.identity.register.resolver import ResolverClient


class ResolverSerializer:
    @staticmethod
    def get_valid_doc_from_token(token: str, get_controller_doc: GetControllerDocFunc) -> RegisterDocument:
        """
        Get a valid RegisterDocument from a resolver token.
        :param token: resolver token
        :param get_controller_doc: get controller register document function
        :return: valid register document

        :raises:
            IdentityResolverError: if invalid token
            IdentityResolverError: if invalid document
        """
        try:
            unverified = JwtTokenHelper.decode_token(token)
            doc = RegisterDocumentBuilder().build_from_dict(unverified['doc'])
            issuer = Issuer.from_string(unverified['iss'])
            issuer_key = RegisterDocumentHelper.get_valid_issuer_key_for_control_only(doc, issuer.name,
                                                                                      get_controller_doc)
            if not issuer_key:
                raise IdentityInvalidRegisterIssuerError(f'Invalid issuer {issuer}')
            JwtTokenHelper.decode_and_verify_token(token, issuer_key.public_key_base58, unverified['aud'])
            return doc
        except (KeyError, ValueError, IdentityValidationError, IdentityInvalidRegisterIssuerError) as exc:
            raise IdentityResolverError(f'Can not deserialized invalid resolver token: \'{exc}\'') from exc

    @staticmethod
    def serialize_to_token(document: RegisterDocument, private_key: ec.EllipticCurvePrivateKey,
                           issuer: Issuer, audience: str = '') -> str:
        """
        Serialize a register document to resolver token.
        :param document: register document
        :param private_key: token issuer private key
        :param issuer: token issuer
        :param audience: token audience
        :return: resolver token

        :raises:
            IdentityResolverError: if can not encode the token
        """
        try:
            return JwtTokenHelper.create_doc_token(issuer, audience, document, private_key)
        except IdentityValidationError as err:
            raise IdentityResolverError(f'Can not serialized to register document resolver token: \'{err}\'') from err


class RESTResolverRequester:

    def __init__(self, address: str, timeout: Optional[Union[int, float]] = 60.0):
        """
        Rest resolver requester.
        :param address: http REST resolver url
        :param timeout: optional timeout seconds. Default=60s. If set to 0, requests will have no timeout.
        """
        self.address = address
        self.timeout = None if timeout and timeout <= 0 else timeout

    def get_token(self, doc_id: str) -> str:
        """
        Request the REST resolver to get the token associated to the provided doc identifier.
        :param doc_id: register document decentralised identifier
        :return: resolver token

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverHttpDocNotFoundError: if document not found
            IdentityResolverHttpError: if http error
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error

        """
        try:
            rsp = requests.get(
                f'{self.address}/1.0/discover/{doc_id}',
                timeout=self.timeout
            )
            rsp.raise_for_status()
            return rsp.json()['token']
        except (KeyError, ValueError) as exc:
            raise IdentityResolverError(f'Unexpected token format received: \'{exc}\'') from exc
        except requests.HTTPError as exc:
            if exc.response.status_code == HTTPStatus.NOT_FOUND:
                raise IdentityResolverHttpDocNotFoundError(f'Identity token for {doc_id} not found') from exc
            raise IdentityResolverHttpError('Identity token could not be retrieved') from exc
        except requests.Timeout as exc:
            raise IdentityResolverTimeoutError(f'Token retrieval from {self.address} '
                                               f'with timeout: \'{self.timeout}\' timed out') from exc
        except requests.RequestException as exc:
            raise IdentityResolverCommunicationError(f'Failed to retrieve token from {self.address}') from exc

    def register_token(self, token: str):
        """
        Register a new document token against the REST resolver.
        :param token: document token

        :raises:
            IdentityResolverHttpError: if http error
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
        """
        try:
            rsp = requests.post(
                f'{self.address}/1.0/register',
                headers={'Content-type': 'text/plain'},
                data=token,
                timeout=self.timeout
            )
            rsp.raise_for_status()
        except requests.HTTPError as exc:
            raise IdentityResolverHttpError(f'Can not register Resolver token ot {self.address}: \'{exc}\'') from exc
        except requests.Timeout as exc:
            raise IdentityResolverTimeoutError(f'Token registration with {self.address} '
                                               f'with timeout: \'{self.timeout}\' timed out') from exc
        except requests.RequestException as exc:
            raise IdentityResolverCommunicationError(
                f'Can not register Resolver token ot {self.address}: \'{exc}\'') from exc


class RESTResolverClient(ResolverClient):

    def __init__(self, requester: RESTResolverRequester, serializer: ResolverSerializer):
        self.requester = requester
        self.serializer = serializer

    def get_document(self, doc_id: str) -> RegisterDocument:
        """
        Get a valid register document from the REST resolver.
        :param doc_id: register document decentralised identifier
        :return: valid register document

        :raises:
            IdentityResolverError: if invalid resolver response
            IdentityResolverHttpDocNotFoundError: if document not found
            IdentityResolverHttpError: if http error
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error

        """
        token = self.requester.get_token(doc_id)
        return self.serializer.get_valid_doc_from_token(token, get_controller_doc=self.get_document)

    def register_document(self, document: RegisterDocument, private_key: ec.EllipticCurvePrivateKey,
                          issuer: Issuer, audience: str = ''):
        """
        Register a register document against the REST Resolver.
        :param document: register document
        :param issuer: issuer
        :param private_key: issuer private key
        :param audience: audience

        :raises:
            IdentityResolverHttpError: if http error
            IdentityResolverTimeoutError: if timeout error
            IdentityResolverCommunicationError: if communication error
            IdentityResolverError: if can not serialize document
        """
        token = self.serializer.serialize_to_token(document, private_key, issuer, audience)
        self.requester.register_token(token)


def get_rest_resolver_client(address: str, timeout: Optional[Union[int, float]] = 60.0) -> RESTResolverClient:
    """
    Get a REST resolver client
    :param address: http REST resolver url
    :param timeout: optional timeout seconds. Default=60s. If set to 0, requests will have no timeout.
    :return: REST resolver client
    """
    requester = RESTResolverRequester(address, timeout)
    serializer = ResolverSerializer()
    return RESTResolverClient(requester, serializer)
